import numpy as np
import torch


def to_gpu(x: torch.Tensor) -> torch.Tensor:
    return x.contiguous().cuda(non_blocking=True)


def audio_float_to_int16(
    audio: np.ndarray, max_wav_value: float = 32767.0
) -> np.ndarray:
    """Normalize audio and convert to int16 range"""
    audio_norm = audio * (max_wav_value / max(0.01, np.max(np.abs(audio))))
    audio_norm = np.clip(audio_norm, -max_wav_value, max_wav_value)
    audio_norm = audio_norm.astype("int16")
    return audio_norm



import os
import requests


class PhonikudDiacritizer:
    dl_url = "https://huggingface.co/thewh1teagle/phonikud-onnx/resolve/main/phonikud-1.0.int8.onnx"

    def __init__(self):

        base_path = os.path.expanduser("~/.local/share/phonikud")
        fname = self.dl_url.split("/")[-1]
        model = f"{base_path}/{fname}"
        if not os.path.isfile(model):
            os.makedirs(base_path, exist_ok=True)
            # TODO - streaming download
            data = requests.get(self.dl_url).content
            with open(model, "wb") as f:
                f.write(data)

        from phonikud_onnx import Phonikud
        self.phonikud = Phonikud(model)

    def diacritize(self, text: str) -> str:
        return self.phonikud.add_diacritics(text)


from typing import Union

from phoonnx.phonemizers.base import BasePhonemizer, UnicodeCodepointPhonemizer, RawPhonemes, GraphemePhonemizer, TextChunks, RawPhonemizedChunks
from phoonnx.phonemizers.en import DeepPhonemizer, OpenPhonemizer, G2PEnPhonemizer
from phoonnx.phonemizers.gl import CotoviaPhonemizer
from phoonnx.phonemizers.vi import VIPhonemePhonemizer
from phoonnx.phonemizers.he import PhonikudPhonemizer
from phoonnx.phonemizers.ar import MantoqPhonemizer
from phoonnx.phonemizers.fa import PersianPhonemizer
from phoonnx.phonemizers.ja import PyKakasiPhonemizer, CutletPhonemizer, OpenJTaklPhonemizer
from phoonnx.phonemizers.ko import KoG2PPhonemizer, G2PKPhonemizer
from phoonnx.phonemizers.zh import (G2pCPhonemizer, G2pMPhonemizer, PypinyinPhonemizer,
                                    XpinyinPhonemizer, JiebaPhonemizer)
from phoonnx.phonemizers.mul import (EspeakPhonemizer, EpitranPhonemizer, MisakiPhonemizer, GoruutPhonemizer,
                                     GruutPhonemizer, ByT5Phonemizer, CharsiuPhonemizer, TransphonePhonemizer)
from phoonnx.phonemizers.mwl import MirandesePhonemizer

Phonemizer = Union[
    MisakiPhonemizer,
    ByT5Phonemizer,
    UnicodeCodepointPhonemizer,
    CharsiuPhonemizer,
    EspeakPhonemizer,
    GruutPhonemizer,
    GoruutPhonemizer,
    EpitranPhonemizer,
    TransphonePhonemizer,
    MirandesePhonemizer,
    OpenJTaklPhonemizer,
    CutletPhonemizer,
    PyKakasiPhonemizer,
    PersianPhonemizer,
    VIPhonemePhonemizer,
    G2PKPhonemizer,
    KoG2PPhonemizer,
    G2pCPhonemizer,
    G2pMPhonemizer,
    PypinyinPhonemizer,
    XpinyinPhonemizer,
    JiebaPhonemizer,
    PhonikudPhonemizer,
    CotoviaPhonemizer,
    MantoqPhonemizer,
    GraphemePhonemizer,
    RawPhonemes,
    OpenPhonemizer,
    G2PEnPhonemizer,
    DeepPhonemizer
]



import datetime
import logging
import re
import string
from datetime import date

from ovos_date_parser import nice_time, nice_date
from ovos_number_parser import pronounce_number, is_fractional, pronounce_fraction
from ovos_number_parser.util import is_numeric
from unicode_rbnf import RbnfEngine, FormatPurpose

LOG = logging.getLogger("normalize")

# A dictionary of common contractions and their expanded forms.
# This list is very comprehensive for English.
CONTRACTIONS = {
    "en": {
        "I'd": "I would",
        "I'll": "I will",
        "I'm": "I am",
        "I've": "I have",
        "ain't": "is not",
        "aren't": "are not",
        "can't": "can not",
        "could've": "could have",
        "couldn't": "could not",
        "didn't": "did not",
        "doesn't": "does not",
        "don't": "do not",
        "gonna": "going to",
        "gotta": "got to",
        "hadn't": "had not",
        "hasn't": "has not",
        "haven't": "have not",
        "he'd": "he would",
        "he'll": "he will",
        "he's": "he is",
        "how'd": "how did",
        "how'll": "how will",
        "how's": "how is",
        "isn't": "is not",
        "it'd": "it would",
        "it'll": "it will",
        "it's": "it is",
        "might've": "might have",
        "mightn't": "might not",
        "must've": "must have",
        "mustn't": "must not",
        "needn't": "need not",
        "oughtn't": "ought not",
        "shan't": "shall not",
        "she'd": "she would",
        "she'll": "she will",
        "she's": "she is",
        "should've": "should have",
        "shouldn't": "should not",
        "somebody's": "somebody is",
        "someone'd": "someone would",
        "someone'll": "someone will",
        "someone's": "someone is",
        "that'd": "that would",
        "that'll": "that will",
        "that's": "that is",
        "there'd": "there would",
        "there're": "there are",
        "there's": "there is",
        "they'd": "they would",
        "they'll": "they will",
        "they're": "they are",
        "they've": "they have",
        "wasn't": "was not",
        "we'd": "we would",
        "we'll": "we will",
        "we're": "we are",
        "we've": "we have",
        "weren't": "were not",
        "what'd": "what did",
        "what'll": "what will",
        "what're": "what are",
        "what's": "what is",
        "what've": "what have",
        "whats": "what is",
        "when'd": "when did",
        "when's": "when is",
        "where'd": "where did",
        "where's": "where is",
        "where've": "where have",
        "who'd": "who would",
        "who'd've": "who would have",
        "who'll": "who will",
        "who're": "who are",
        "who's": "who is",
        "who've": "who have",
        "why'd": "why did",
        "why're": "why are",
        "why's": "why is",
        "won't": "will not",
        "won't've": "will not have",
        "would've": "would have",
        "wouldn't": "would not",
        "wouldn't've": "would not have",
        "y'ain't": "you are not",
        "y'aint": "you are not",
        "y'all": "you all",
        "ya'll": "you all",
        "you'd": "you would",
        "you'd've": "you would have",
        "you'll": "you will",
        "you're": "you are",
        "you've": "you have",
        "I'm'a": "I am going to",
        "I'm'o": "I am going to",
        "I'll've": "I will have",
        "I'd've": "I would have",
        "Whatcha": "What are you",
        "amn't": "am not",
        "'cause": "because",
        "can't've": "cannot have",
        "couldn't've": "could not have",
        "daren't": "dare not",
        "daresn't": "dare not",
        "dasn't": "dare not",
        "everyone's": "everyone is",
        "gimme": "give me",
        "gon't": "go not",
        "hadn't've": "had not have",
        "he've": "he would have",
        "he'll've": "he will have",
        "he'd've": "he would have",
        "here's": "here is",
        "how're": "how are",
        "how'd'y": "how do you do",
        "howd'y": "how do you do",
        "howdy": "how do you do",
        "'tis": "it is",
        "'twas": "it was",
        "it'll've": "it will have",
        "it'd've": "it would have",
        "kinda": "kind of",
        "let's": "let us",
        "ma'am": "madam",
        "may've": "may have",
        "mayn't": "may not",
        "mightn't've": "might not have",
        "mustn't've": "must not have",
        "needn't've": "need not have",
        "ol'": "old",
        "oughtn't've": "ought not have",
        "sha'n't": "shall not",
        "shan't": "shall not",
        "shalln't": "shall not",
        "shan't've": "shall not have",
        "she'd've": "she would have",
        "shouldn't've": "should not have",
        "so've": "so have",
        "so's": "so is",
        "something's": "something is",
        "that're": "that are",
        "that'd've": "that would have",
        "there'll": "there will",
        "there'd've": "there would have",
        "these're": "these are",
        "they'll've": "they will have",
        "they'd've": "they would have",
        "this's": "this is",
        "this'll": "this will",
        "this'd": "this would",
        "those're": "those are",
        "to've": "to have",
        "wanna": "want to",
        "we'll've": "we will have",
        "we'd've": "we would have",
        "what'll've": "what will have",
        "when've": "when have",
        "where're": "where are",
        "which's": "which is",
        "who'll've": "who will have",
        "why've": "why have",
        "will've": "will have",
        "y'all're": "you all are",
        "y'all've": "you all have",
        "y'all'd": "you all would",
        "y'all'd've": "you all would have",
        "you'll've": "you will have"
    }
}

# Dictionaries for titles, units, and their full word equivalents.
TITLES = {
    "en": {
        "Dr.": "Doctor",
        "Mr.": "Mister",
        "Prof.": "Professor"
    },
    "ca": {
        "Dr.": "Doctor",
        "Sr.": "Senyor",
        "Sra.": "Senyora",
        "Prof.": "Professor"
    },
    "es": {
        "Dr.": "Doctor",
        "Sr.": "Señor",
        "Sra.": "Señora",
        "Prof.": "Profesor",
        "D.": "Don",
        "Dña.": "Doña"
    },
    "pt": {
        "Dr.": "Doutor",
        "Sr.": "Senhor",
        "Sra.": "Senhora",
        "Prof.": "Professor",
        "Drª.": "Doutora",
        "Eng.": "Engenheiro",
        "D.": "Dom",
        "Dª": "Dona"
    },
    "gl": {
        "Dr.": "Doutor",
        "Sr.": "Señor",
        "Sra.": "Señora",
        "Prof.": "Profesor",
        "Srta.": "Señorita"
    },
    "fr": {
        "Dr.": "Docteur",
        "M.": "Monsieur",
        "Mme": "Madame",
        "Mlle": "Mademoiselle",
        "Prof.": "Professeur",
        "Pr.": "Professeur"
    },
    "it": {
        "Dr.": "Dottore",
        "Sig.": "Signore",
        "Sig.ra": "Signora",
        "Prof.": "Professore",
        "Dott.ssa": "Dottoressa",
        "Sig.na": "Signorina"
    },
    "nl": {
        "Dr.": "Dokter",
        "Dhr.": "De Heer",
        "Mevr.": "Mevrouw",
        "Prof.": "Professor",
        "Drs.": "Dokterandus",
        "Ing.": "Ingenieur"
    },
    "de": {
        "Dr.": "Doktor",
        "Prof.": "Professor"
    }
}

UNITS = {
    "en": {
        "€": "euros",
        "%": "per cent",
        "°C": "degrees celsius",
        "°F": "degrees fahrenheit",
        "°K": "degrees kelvin",
        "°": "degrees",
        "$": "dollars",
        "£": "pounds",
        "km": "kilometers",
        "m": "meters",
        "cm": "centimeters",
        "mm": "millimeters",
        "ft": "feet",
        "in": "inches",
        "yd": "yards",
        "mi": "miles",
        "kg": "kilograms",
        "g": "grams",
        "lb": "pounds",
        "oz": "ounces",
        "L": "liters",
        "mL": "milliliters",
        "gal": "gallons",
        "qt": "quarts",
        "pt": "pints",
        "hr": "hours",
        "min": "minutes",
        "s": "seconds"
    },
    "pt": {
        "€": "euros",
        "%": "por cento",
        "°C": "graus celsius",
        "°F": "graus fahrenheit",
        "°K": "graus kelvin",
        "°": "graus",
        "$": "dólares",
        "£": "libras",
        "km": "quilômetros",
        "m": "metros",
        "cm": "centímetros",
        "mm": "milímetros",
        "kg": "quilogramas",
        "g": "gramas",
        "L": "litros",
        "mL": "mililitros",
        "h": "horas",
        "min": "minutos",
        "s": "segundos"
    },
    "es": {
        "€": "euros",
        "%": "por ciento",
        "°C": "grados celsius",
        "°F": "grados fahrenheit",
        "°K": "grados kelvin",
        "°": "grados",
        "$": "dólares",
        "£": "libras",
        "km": "kilómetros",
        "m": "metros",
        "cm": "centímetros",
        "kg": "kilogramos",
        "g": "gramos",
        "L": "litros",
        "mL": "millilitros"
    },
    "fr": {
        "€": "euros",
        "%": "pour cent",
        "°C": "degrés celsius",
        "°F": "degrés fahrenheit",
        "°K": "degrés kelvin",
        "°": "degrés",
        "$": "dollars",
        "£": "livres",
        "km": "kilomètres",
        "m": "mètres",
        "cm": "centimètres",
        "kg": "kilogrammes",
        "g": "grammes",
        "L": "litres",
        "mL": "millilitres"
    },
    "de": {
        "€": "Euro",
        "%": "Prozent",
        "°C": "Grad Celsius",
        "°F": "Grad Fahrenheit",
        "°K": "Grad Kelvin",
        "°": "Grad",
        "$": "Dollar",
        "£": "Pfund",
        "km": "Kilometer",
        "m": "Meter",
        "cm": "Zentimeter",
        "kg": "Kilogramm",
        "g": "Gramm",
        "L": "Liter",
        "mL": "Milliliter"
    }
}


def _get_number_separators(full_lang: str) -> tuple[str, str]:
    """
    Determines decimal and thousands separators based on language.
    Defaults to '.' decimal and ',' thousands for most languages.
    Special cases:
    - 'pt', 'es', 'fr', 'de': ',' decimal and '.' thousands.
    """
    lang_code = full_lang.split("-")[0]
    decimal_separator = '.'
    thousands_separator = ','
    if lang_code in ["pt", "es", "fr", "de"]:
        decimal_separator = ','
        thousands_separator = '.'
    return decimal_separator, thousands_separator


def _normalize_number_word(word: str, full_lang: str, rbnf_engine) -> str:
    """
    Helper function to normalize a single word that is a number, handling
    decimal and thousands separators based on locale.
    """
    cleaned_word = word.rstrip(string.punctuation)

    # Handle fractions like '3/3'
    if is_fraction(cleaned_word):
        try:
            return pronounce_fraction(cleaned_word, full_lang) + word[len(cleaned_word):]
        except Exception as e:
            LOG.error(f"ovos-number-parser failed to pronounce fraction: {word} - ({e})")
            return word

    # Handle numbers with locale-specific separators
    decimal_separator, thousands_separator = _get_number_separators(full_lang)
    temp_cleaned_word = cleaned_word

    # Check if the word contains a thousands separator followed by digits and a decimal separator
    # This is a specific check for formats like '123.456,78'
    has_thousands_and_decimal = (
            thousands_separator in temp_cleaned_word and
            decimal_separator in temp_cleaned_word and
            temp_cleaned_word.index(thousands_separator) < temp_cleaned_word.index(decimal_separator)
    )

    if has_thousands_and_decimal:
        temp_cleaned_word = temp_cleaned_word.replace(thousands_separator, "")
        temp_cleaned_word = temp_cleaned_word.replace(decimal_separator, ".")
    elif decimal_separator in temp_cleaned_word and is_numeric(temp_cleaned_word.replace(decimal_separator, ".", 1)):
        # Handle cases like '1,2' -> '1.2'
        temp_cleaned_word = temp_cleaned_word.replace(decimal_separator, ".")
    elif thousands_separator in temp_cleaned_word and is_numeric(temp_cleaned_word.replace(thousands_separator, "", 1)):
        # Handle cases like '1.234' -> '1234'
        temp_cleaned_word = temp_cleaned_word.replace(thousands_separator, "")

    # Check if the word is a valid number after processing
    if is_numeric(temp_cleaned_word):
        try:
            num = float(temp_cleaned_word) if "." in temp_cleaned_word else int(temp_cleaned_word)
            return pronounce_number(num, lang=full_lang) + word[len(cleaned_word):]
        except Exception as e:
            LOG.error(f"ovos-number-parser failed to pronounce number: {word} - ({e})")
            return word

    elif rbnf_engine and cleaned_word.isdigit():
        try:
            pronounced_number = rbnf_engine.format_number(cleaned_word, FormatPurpose.CARDINAL).text
            return pronounced_number + word[len(cleaned_word):]
        except Exception as e:
            LOG.error(f"unicode-rbnf failed to pronounce number: {word} - ({e})")
            return word

    return word


# --- Date and Time Pronunciation ---
def pronounce_date(date_obj: date, full_lang: str) -> str:
    """
    Pronounces a date object using ovos-date-parser.
    """
    return nice_date(date_obj, full_lang)


def pronounce_time(time_string: str, full_lang: str) -> str:
    """
    Pronounces a time string using ovos-date-parser.
    Handles military time like "15h01" and converts it to a
    datetime.time object before passing it to nice_time.
    """
    try:
        hours, mins = time_string.split("h")
        time_obj = datetime.time(int(hours), int(mins))
        # Use nice_time from ovos-date-parser
        return nice_time(time_obj, full_lang, speech=True, use_24hour=True, use_ampm=False)
    except Exception as e:
        LOG.warning(f"Failed to parse time string '{time_string}': {e}")
        return time_string.replace("h", " ")


def _normalize_dates_and_times(text: str, full_lang: str, date_format: str = "DMY") -> str:
    """
    Helper function to normalize dates and times using regular expressions.
    This prepares the strings for pronunciation.
    """
    lang_code = full_lang.split("-")[0]
    # Pre-process with regex to handle English am/pm times
    if lang_code == "en":
        text = re.sub(r"(?i)(\d+)(am|pm)", r"\1 \2", text)
        # Handle the pronunciation for TTS
        text = text.replace("am", "A M").replace("pm", "P M")

    # Normalize times like "15h01" to words
    time_pattern = re.compile(r"(\d{1,2})h(\d{2})", re.IGNORECASE)

    def replace_time(match):
        time_str = match.group(0)
        return pronounce_time(time_str, full_lang)

    text = time_pattern.sub(replace_time, text)

    # Find dates like "DD/MM/YYYY" or "YYYY/MM/DD"
    date_pattern = re.compile(r"(\d{1,4})[/-](\d{1,2})[/-](\d{1,4})")

    match = date_pattern.search(text)

    if match:
        # Get the three parts of the date string
        part1_str, part2_str, part3_str = match.groups()
        p1, p2, p3 = int(part1_str), int(part2_str), int(part3_str)

        # Initialize month, day, and year
        month, day, year = None, None, None

        # Determine year first based on length (4 digits)
        if len(part1_str) == 4:
            year, rest_parts = p1, [p2, p3]
        elif len(part3_str) == 4:
            year, rest_parts = p3, [p1, p2]
        else:
            # If no 4-digit year, it's ambiguous, assume a 2-digit year.
            # We'll assume the last part is the year based on common patterns.
            year = p3
            # Expand 2-digit year to 4-digit year
            if year < 100:
                # Assume years 00-29 are 2000-2029, 30-99 are 1930-1999
                year = 2000 + year if year < 30 else 1900 + year
            rest_parts = [p1, p2]

        # From the remaining parts, try to determine day and month
        if day is None and any(p > 12 and len(str(p)) == 2 for p in rest_parts):
            # If a two-digit number is > 12, it's a day
            day_candidate = next((p for p in rest_parts if p > 12), None)
            if day_candidate:
                day = day_candidate
                rest_parts.remove(day_candidate)
                month = rest_parts[0]

        # Fallback to date_format if day/month are still ambiguous
        if day is None or month is None:
            if date_format.lower() == "mdy":
                month, day = rest_parts[0], rest_parts[1]
            else:  # default to DD/MM/YY
                day, month = rest_parts[0], rest_parts[1]

        try:
            date_obj = date(year, month, day)
            pronounced_date_str = pronounce_date(date_obj, full_lang)
            text = text.replace(match.group(0), pronounced_date_str)
        except (ValueError, IndexError) as e:
            LOG.warning(f"Could not parse date from '{match.group(0)}': {e}")

    return text


def _normalize_word_hyphen_digit(text: str) -> str:
    """
    Helper function to normalize words attached to digits with a hyphen,
    such as 'sub-23' -> 'sub 23'.
    """
    # Regex to find a word (\w+) followed by a hyphen and a digit (\d+)
    pattern = re.compile(r"(\w+)-(\d+)")
    text = pattern.sub(r"\1 \2", text)
    return text


def _normalize_units(text: str, full_lang: str) -> str:
    """
    Helper function to normalize units attached to numbers.
    This function handles symbolic and alphanumeric units separately
    to avoid issues with word boundaries.
    """
    text = text.replace("º", "°") # these characters look the same... but...
    lang_code = full_lang.split("-")[0]
    if lang_code in UNITS:
        # Determine number separators for the language
        decimal_separator, thousands_separator = _get_number_separators(full_lang)

        # Separate units into symbolic and alphanumeric
        symbolic_units = {k: v for k, v in UNITS[lang_code].items() if not k.isalnum()}
        alphanumeric_units = {k: v for k, v in UNITS[lang_code].items() if k.isalnum()}

        # Create regex pattern for symbolic units and replace them first
        sorted_symbolic = sorted(symbolic_units.keys(), key=len, reverse=True)
        symbolic_pattern_str = "|".join(re.escape(unit) for unit in sorted_symbolic)
        if symbolic_pattern_str:
            # Pattern to match numbers with optional thousands and decimal separators
            number_pattern_str = rf"(\d+[{re.escape(thousands_separator)}]?\d*[{re.escape(decimal_separator)}]?\d*)"
            symbolic_pattern = re.compile(number_pattern_str + r"\s*(" + symbolic_pattern_str + r")", re.IGNORECASE)

            def replace_symbolic(match):
                number = match.group(1)
                # Remove thousands separator and replace decimal separator for parsing
                if thousands_separator in number and decimal_separator in number:
                    number = number.replace(thousands_separator, "").replace(decimal_separator, ".")
                elif decimal_separator != "." and decimal_separator in number:
                    number = number.replace(decimal_separator, ".")
                unit_symbol = match.group(2)
                unit_word = symbolic_units[unit_symbol]
                try:
                    return f"{pronounce_number(float(number) if '.' in number else int(number), full_lang)} {unit_word}"
                except Exception as e:
                    LOG.error(f"Failed to pronounce number with unit: {number}{unit_symbol} - ({e})")
                    return match.group(0)
            text = symbolic_pattern.sub(replace_symbolic, text)

        # Create regex pattern for alphanumeric units and replace them next
        sorted_alphanumeric = sorted(alphanumeric_units.keys(), key=len, reverse=True)
        alphanumeric_pattern_str = "|".join(re.escape(unit) for unit in sorted_alphanumeric)
        if alphanumeric_pattern_str:
            number_pattern_str = rf"(\d+[{re.escape(thousands_separator)}]?\d*[{re.escape(decimal_separator)}]?\d*)"
            alphanumeric_pattern = re.compile(number_pattern_str + r"\s*(" + alphanumeric_pattern_str + r")\b",
                                              re.IGNORECASE)

            def replace_alphanumeric(match):
                number = match.group(1)
                # Remove thousands separator and replace decimal separator for parsing
                if thousands_separator in number and decimal_separator in number:
                    number = number.replace(thousands_separator, "").replace(decimal_separator, ".")
                elif decimal_separator != "." and decimal_separator in number:
                    number = number.replace(decimal_separator, ".")
                unit_symbol = match.group(2)
                unit_word = alphanumeric_units[unit_symbol]
                return f"{pronounce_number(float(number) if '.' in number else int(number), full_lang)} {unit_word}"

            text = alphanumeric_pattern.sub(replace_alphanumeric, text)
    return text


def _normalize_word(word: str, full_lang: str, rbnf_engine) -> str:
    """
    Helper function to normalize a single word.
    """
    lang_code = full_lang.split("-")[0]

    if word in CONTRACTIONS.get(lang_code, {}):
        return CONTRACTIONS[lang_code][word]

    if word in TITLES.get(lang_code, {}):
        return TITLES[lang_code][word]

    # Delegate number parsing to the new helper function
    normalized_number = _normalize_number_word(word, full_lang, rbnf_engine)
    if normalized_number != word:
        return normalized_number

    return word


def is_fraction(word: str) -> bool:
    """Checks if a word is a fraction like '3/3'."""
    if "/" in word:
        parts = word.split("/")
        if len(parts) == 2:
            n1, n2 = parts
            return n1.isdigit() and n2.isdigit()
    return False


def normalize(text: str, lang: str) -> str:
    """
    Normalizes a text string by expanding contractions, titles, and pronouncing
    numbers, units, and fractions.
    """
    full_lang = lang
    lang_code = full_lang.split("-")[0]
    dialog = text

    # Step 1: Handle dates and times with ovos-date-parser
    date_format = "MDY" if full_lang.lower() == "en-us" else "DMY"
    dialog = _normalize_dates_and_times(dialog, full_lang, date_format)

    # Step 2: Normalize words with hyphens and digits
    dialog = _normalize_word_hyphen_digit(dialog)

    # Step 3: Expand units attached to numbers
    dialog = _normalize_units(dialog, full_lang)

    # Step 4: Normalize word-by-word
    words = dialog.split()
    rbnf_engine = None
    try:
        rbnf_engine = RbnfEngine.for_language(lang_code)
    except (ValueError, KeyError) as e:
        LOG.debug(f"RBNF engine not available for language '{lang_code}': {e}")

    normalized_words = [_normalize_word(word, full_lang, rbnf_engine) for word in words]
    dialog = " ".join(normalized_words)

    return dialog


if __name__ == "__main__":
    # --- Example usage for demonstration purposes ---

    # General normalization examples
    print("General English example: " + normalize('I\'m Dr. Prof. 3/3 0.5% of 12345€, 5ft, and 10kg', 'en'))
    print(f"Word Salad Portuguese (Dr. Prof. 3/3 0,5% de 12345€, 5m, e 10kg): {normalize('Dr. Prof. 3/3 0,5% de 12345€, 5m, e 10kg', 'pt')}")
    print(f"Word Salad Portuguese (Dr. Prof. 3/3 0.5% de 12345€, 5m, e 10kg): {normalize('Dr. Prof. 3/3 0.5% de 12345€, 5m, e 10kg', 'pt')}")

    # Portuguese examples with comma decimal separator
    print("\n--- Portuguese Decimal Separator Examples ---")
    print(
        f"Original: 'A coima aplicada é de 1,2 milhões de euros.' Normalized: '{normalize('A coima aplicada é de 1,2 milhões de euros.', 'pt')}'")
    print(
        f"Original: 'Agora, tem 1,88 metros e muito para contar.' Normalized: '{normalize('Agora, tem 1,88 metros e muito para contar.', 'pt')}'")
    print(
        f"Original: 'Ainda temos 1,7 milhões de pobres!' Normalized: '{normalize('Ainda temos 1,7 milhões de pobres!', 'pt')}'")
    print(f"Original: 'O lucro foi de 123.456,78€.' Normalized: '{normalize('O lucro foi de 123.456,78€.', 'pt')}'")
    print(f"Normalized: '{normalize('O lucro foi de 123.456,78€.', 'pt-PT')}'")

    # English dates and times
    print("\n--- English Date & Time Examples ---")
    print(f"English date (MDY format): {normalize('The date is 08/03/2025', 'en-US')}")
    print(f"English ambiguous date (MDY assumed): {normalize('The report is due 15/05/2025', 'en-US')}")
    print(f"English date with dashes: {normalize('The event is on 11-04-2025', 'en-US')}")
    print(f"English AM/PM time: {normalize('The meeting is at 10am', 'en-US')}")
    print(f"English military time: {normalize('The party is at 19h30', 'en-US')}")
    print(f"English month name: {normalize('The report is due 15 May 2025', 'en-US')}")

    # Portuguese dates and times
    print("\n--- Portuguese Date & Time Examples ---")
    print(f"Portuguese date (A data é 03/08/2025): {normalize('A data é 03/08/2025', 'pt')}")
    print(f"Portuguese ambiguous date (O relatório é para 15/05/2025): {normalize('O relatório é para 15/05/2025', 'pt')}")
    print(f"Portuguese date with dashes (O evento é no dia 25-10-2024): {normalize('O evento é no dia 25-10-2024', 'pt')}")
    print(f"Portuguese military time (O encontro é às 14h30): {normalize('O encontro é às 14h30', 'pt')}")

    # Other examples
    print(f"\n--- Other Examples ---")
    print(f"English fraction: {normalize('The fraction is 1/2', 'en')}")
    print(f"English plural fraction: {normalize('There are 3/4 of a cup', 'en')}")
    print(f"Spanish example with units: {normalize('The temperature is 25ºC', 'es')}")
    print(f"Portuguese with punctuation: {normalize('12345€, 5m e 10kg', 'pt')}")
    print(
        f"Portuguese word-digit: {normalize('Esta temporada leva oito jogos ao serviço da equipa sub-23 leonina.', 'pt')}")



from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.config import Alphabet


class VIPhonemePhonemizer(BasePhonemizer):
    """https://github.com/v-nhandt21/Viphoneme"""
    def __init__(self):
        from viphoneme import vi2IPA
        self.g2p = vi2IPA
        super().__init__(Alphabet.IPA)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["vi"])

    def phonemize_string(self, text: str, lang: str = "vi") -> str:
        """
        """
        lang = self.get_lang(lang)
        return self.g2p(text)


if __name__ == "__main__":
    text = "Được viết vào 6/4/2020, có thể xử lí những trường hợp chứa English"

    pho = VIPhonemePhonemizer()
    lang = "vi"

    print(f"\n--- Getting phonemes for '{text}' ---")
    phonemes_cotovia = pho.phonemize(text, lang)
    print(f"  Phonemes: {phonemes_cotovia}")



import abc
from typing import List


from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.thirdparty.zh_num import num2str
from phoonnx.config import Alphabet


class JiebaPhonemizer(BasePhonemizer):
    """
    A non-phonemizing class that simply uses Jieba to segment Chinese text
    into words with spaces for token separation.
    """
    def __init__(self):
        super().__init__(Alphabet.HANZI)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["zh"])

    def phonemize_string(self, text: str, lang: str = "zh") -> str:
        """
        Segments the input Chinese text using Jieba.

        Args:
            text (str): The input sentence.
            lang (str): Language code (must be "zh").

        Returns:
            str: Tokenized text with words separated by spaces.
        """
        import jieba
        lang = self.get_lang(lang)
        seg_list = jieba.cut(text, cut_all=False)
        seg_list = [num2str(w) if w.isdigit() else w for w in seg_list]
        return " ".join(seg_list)


class BaseChinesePinyinPhonemizer(BasePhonemizer):
    """
    Base class for Chinese phonemizers using different pinyin G2P libraries.
    Supports optional IPA conversion and segmentation via Jieba.
    """

    def __init__(self, alphabet=Alphabet.PINYIN, jieba: bool = True, retone=True):
        """
        Initializes the phonemizer.

        Args:
            ipa (bool): Whether to convert pinyin to IPA.
            jieba (bool): Whether to segment text using Jieba before phonemization.
        """
        assert alphabet in [Alphabet.PINYIN, Alphabet.IPA]
        super().__init__(alphabet)
        self.jieba = jieba
        self.retone = retone
        from pinyin_to_ipa import pinyin_to_ipa
        self.pinyin_to_ipa = pinyin_to_ipa

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["zh"])

    @staticmethod
    def _retone(p):
        p = p.replace('˧˩˧', '↓')  # third tone
        p = p.replace('˧˥', '↗')  # second tone
        p = p.replace('˥˩', '↘')  # fourth tone
        p = p.replace('˥', '→')  # first tone
        p = p.replace(chr(635) + chr(809), 'ɨ').replace(chr(633) + chr(809), 'ɨ')
        assert chr(809) not in p, p
        return p

    def to_ipa(self, phones: List[str]) -> List[str]:
        """
        Converts a list of pinyin syllables to IPA. Falls back to the original syllable if conversion fails.

        Args:
            phones (List[str]): List of pinyin syllables or phrases.

        Returns:
            List[str]: Corresponding IPA or original syllables.
        """
        ipa_phones: List[str] = []
        for p in phones:
            if p == " ":
                ipa_phones.append(" ")
                continue
            pho_str = ""
            for sp in p.split():  # G2P might return phrases with multiple syllables
                try:
                    pho = self.pinyin_to_ipa(sp.strip())[0][0]
                    if self.retone:
                        pho = self._retone(pho)
                    pho_str += pho
                except Exception:
                    pass
            ipa_phones.append(pho_str)
        return ipa_phones

    def phonemize_to_list(self, text: str, lang: str) -> List[str]:
        phones: List[str] = []
        lang = self.get_lang(lang)
        if self.jieba:
            import jieba
            for chunk in jieba.cut(text, cut_all=False):
                if chunk.isdigit():
                    chunk = num2str(chunk)
                phones += self.get_pinyin(chunk)
                phones += [" "]  # keep jieba whitespace
        else:
            phones = self.get_pinyin(text)
        if self.alphabet == Alphabet.IPA:
            phones = self.to_ipa(phones)
        return phones

    def phonemize_string(self, text: str, lang: str = "zh") -> str:
        """
        Converts input text to a space-separated phoneme string.

        Args:
            text (str): The input sentence.
            lang (str): Language code (must be "zh").

        Returns:
            str: Space-separated phoneme string.
        """
        phones: List[str] = self.phonemize_to_list(text, lang)
        return "".join(phones)

    @abc.abstractmethod
    def get_pinyin(self, text: str) -> List[str]:
        """
        Abstract method to be implemented by subclasses for converting text to pinyin.

        Args:
            text (str): Input Chinese text.

        Returns:
            List[str]: List of pinyin tokens.
        """
        return NotImplemented


class G2pCPhonemizer(BaseChinesePinyinPhonemizer):
    """
    Phonemizer using g2pc (CRF-based Grapheme-to-Phoneme converter).
    https://github.com/Kyubyong/g2pC
    """

    def __init__(self, alphabet=Alphabet.PINYIN, jieba: bool = True):
        from g2pc import G2pC
        self.g2p = G2pC()
        super().__init__(alphabet, jieba)

    def get_pinyin(self, text: str) -> List[str]:
        """
        Returns a list of pinyin syllables from g2pc.

        Args:
            text (str): Input Chinese text.

        Returns:
            List[str]: Pinyin tokens.
        """
        return [a[3] for a in self.g2p(text)]


class G2pMPhonemizer(BaseChinesePinyinPhonemizer):
    """
    Phonemizer using g2pM - A Neural Grapheme-to-Phoneme Conversion Package for Mandarin Chinese
    https://github.com/kakaobrain/g2pm
    """

    def __init__(self, alphabet=Alphabet.PINYIN, tone: bool = True, char_split: bool = False, jieba: bool = True):
        from g2pM import G2pM
        self.g2p = G2pM()
        self.tone = tone
        self.char_split = char_split
        super().__init__(alphabet, jieba)

    def get_pinyin(self, text: str) -> List[str]:
        """
        Returns a list of pinyin tokens from g2pM.

        Args:
            text (str): Input Chinese text.

        Returns:
            List[str]: Pinyin tokens.
        """
        return self.g2p(text, tone=self.tone, char_split=self.char_split)


class XpinyinPhonemizer(BaseChinesePinyinPhonemizer):
    """
    Phonemizer using xpinyin (basic pinyin generator with optional tone marks).
    """

    def __init__(self, alphabet=Alphabet.PINYIN, tone_marks: str = "numbers", jieba: bool = True):
        from xpinyin import Pinyin
        self.g2p = Pinyin()
        self.tone_marks = tone_marks
        super().__init__(alphabet, jieba)

    def get_pinyin(self, text: str) -> List[str]:
        """
        Returns a list of pinyin tokens from xpinyin.

        Args:
            text (str): Input Chinese text.

        Returns:
            List[str]: Pinyin tokens.
        """
        return self.g2p.get_pinyin(text, tone_marks=self.tone_marks).split("-")


class PypinyinPhonemizer(BaseChinesePinyinPhonemizer):
    """
    Phonemizer using pypinyin (comprehensive and accurate pinyin library).
    """

    def __init__(self, alphabet=Alphabet.PINYIN, jieba: bool = True):
        from pypinyin import pinyin
        self.g2p = pinyin
        super().__init__(alphabet, jieba)

    def get_pinyin(self, text: str) -> List[str]:
        """
        Returns a list of pinyin tokens from pypinyin.

        Args:
            text (str): Input Chinese text.

        Returns:
            List[str]: Pinyin tokens.
        """
        return [p[0] for p in self.g2p(text)]


if __name__ == "__main__":
    lang = "zh"
    text = "然而，他红了20年以后，他竟退出了大家的视线。"

    pho = JiebaPhonemizer()
    #pho1 = G2pCPhonemizer(ipa=True)
    pho2 = G2pMPhonemizer()
    pho3 = XpinyinPhonemizer()
    pho4 = PypinyinPhonemizer()

    from phoonnx.phonemizers.mul import MisakiPhonemizer

    pho5 = MisakiPhonemizer()

    print(f"\n--- Getting phonemes for '{text}' ---")

    phones = pho5.phonemize_to_list(text, lang)
    print(f" Misaki: {phones}")

    #phones = pho1.phonemize(text, lang)
    #print(f" G2pC: {phones}")

    phones = pho2.phonemize_to_list(text, lang)
    print(f" G2pM: {phones}")

    phones = pho3.phonemize_to_list(text, lang)
    print(f" Xpinyin: {phones}")

    phones = pho4.phonemize_to_list(text, lang)
    print(f" Pypinyin: {phones}")

    phones = pho.phonemize_to_list(text, lang)
    print(f" Jieba: {phones}")

    #
    exit()
    #   Phonemes: [('ran2 er2 ， ta1 hong2 le5 2 0 nian2 yi3 hou4 ， ta1 jing4 tui4 chu1 le5 da4 jia1 de5 shi4 xian4 。', '.', True)]
    #   Phonemes: [('ran2 er2 ， ta1 hong2 le5 20 nian2 yi3 hou4 ， ta1 jing4 tui4 chu1 le5 da4 jia1 de5 shi4 xian4 。', '.', True)]
    #   Phonemes: [('ran2 er2 ， ta1 hong2 le5 20 nian2 yi3 hou4 ， ta1 jing4 tui4 chu1 le5 da4 jia1 de5 shi4 xian4 。', '.', True)]
    #   Phonemes: [('rán ér ， tā hóng le 20 nián yǐ hòu ， tā jìng tuì chū le dà jiā de shì xiàn 。', '.', True)]



import json
import logging
from pathlib import Path
import os
import torch
import click
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint

from phoonnx_train.vits.lightning import VitsModel

_LOGGER = logging.getLogger(__package__)


def load_state_dict(model, saved_state_dict):
    state_dict = model.state_dict()
    new_state_dict = {}

    for k, v in state_dict.items():
        if k in saved_state_dict:
            new_state_dict[k] = saved_state_dict[k]
        else:
            _LOGGER.debug("%s is not in the checkpoint", k)
            new_state_dict[k] = v

    model.load_state_dict(new_state_dict)


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option('--dataset-dir', required=True, type=click.Path(exists=True, file_okay=False), help='Path to pre-processed dataset directory')
@click.option('--checkpoint-epochs', default=1, type=int, help='Save checkpoint every N epochs (default: 1)')
@click.option('--quality', default='medium', type=click.Choice(['x-low', 'medium', 'high']), help='Quality/size of model (default: medium)')
@click.option('--resume-from-checkpoint', default=None, help='Load an existing checkpoint and resume training')
@click.option('--resume-from-single-speaker-checkpoint', help='For multi-speaker models only. Converts a single-speaker checkpoint to multi-speaker and resumes training')
@click.option('--seed', type=int, default=1234, help='Random seed (default: 1234)')
# Common Trainer options
@click.option('--max-epochs', type=int, default=1000, help='Stop training once this number of epochs is reached (default: 1000)')
@click.option('--devices', default=1, help='Number of devices or list of device IDs to train on (default: 1)')
@click.option('--accelerator', default='auto', help='Hardware accelerator to use (cpu, gpu, tpu, mps, etc.)  (default: "auto")')
@click.option('--default-root-dir', type=click.Path(file_okay=False), default=None, help='Default root directory for logs and checkpoints (default: None)')
@click.option('--precision', default=32, help='Precision used in training (e.g. 16, 32, bf16) (default: 32)')
# Model-specific arguments
@click.option('--learning-rate', type=float, default=2e-4, help='Learning rate for optimizer (default: 2e-4)')
@click.option('--batch-size', type=int, default=16, help='Training batch size (default: 16)')
@click.option('--num-workers', type=click.IntRange(min=1), default=os.cpu_count() or 1, help='Number of data loader workers (default: CPU count)')
@click.option('--validation-split', type=float, default=0.05, help='Proportion of data used for validation (default: 0.05)')
@click.option('--discard-encoder', type=bool, default=False, help='Discard the encoder weights from base checkpoint (default: False)')
def main(
    dataset_dir,
    checkpoint_epochs,
    quality,
    resume_from_checkpoint,
    resume_from_single_speaker_checkpoint,
    seed,
    max_epochs,
    devices,
    accelerator,
    default_root_dir,
    precision,
    learning_rate,
    batch_size,
    num_workers,
    validation_split,
    discard_encoder
):
    logging.basicConfig(level=logging.DEBUG)

    dataset_dir = Path(dataset_dir)
    if default_root_dir is None:
        default_root_dir = dataset_dir

    torch.backends.cudnn.benchmark = True
    torch.manual_seed(seed)

    config_path = dataset_dir / 'config.json'
    dataset_path = dataset_dir / 'dataset.jsonl'

    _LOGGER.info(f"config_path: '{config_path}'")
    _LOGGER.info(f"dataset_path: '{dataset_path}'")

    with open(config_path, 'r', encoding='utf-8') as config_file:
        config = json.load(config_file)

    trainer = Trainer(
        max_epochs=max_epochs,
        devices=devices,
        accelerator=accelerator,
        default_root_dir=default_root_dir,
        precision=precision
    )

    if checkpoint_epochs is not None:
        trainer.callbacks = [ModelCheckpoint(every_n_epochs=checkpoint_epochs)]
        _LOGGER.info('Checkpoints will be saved every %s epoch(s)', checkpoint_epochs)

    dict_args = dict(
        seed=seed,
        learning_rate=learning_rate,
        batch_size=batch_size,
        num_workers=num_workers,
        validation_split=validation_split,
    )

    if quality == 'x-low':
        dict_args.update({
            'hidden_channels': 96,
            'inter_channels': 96,
            'filter_channels': 384,
        })
    elif quality == 'high':
        dict_args.update({
            'resblock': '1',
            'resblock_kernel_sizes': (3, 7, 11),
            'resblock_dilation_sizes': ((1, 3, 5), (1, 3, 5), (1, 3, 5)),
            'upsample_rates': (8, 8, 2, 2),
            'upsample_initial_channel': 512,
            'upsample_kernel_sizes': (16, 16, 4, 4),
        })

    num_symbols = int(config['num_symbols'])
    num_speakers = int(config['num_speakers'])
    sample_rate = int(config['audio']['sample_rate'])
    _LOGGER.debug(f"Config params: num_symbols={num_symbols} num_speakers={num_speakers} sample_rate={sample_rate}")

    if resume_from_checkpoint:
        # TODO (?) - add a flag to use params from config vs from checkpoint in case of mismatch
        ckpt = VitsModel.load_from_checkpoint(resume_from_checkpoint, dataset=None)
        _LOGGER.debug(f"Checkpoint params: num_symbols={ckpt.model_g.n_vocab} num_speakers={ckpt.model_g.n_speakers} sample_rate={ckpt.hparams.sample_rate}")
        if ckpt.model_g.n_vocab != num_symbols:
            _LOGGER.warning(f"Checkpoint num_symbols={ckpt.model_g.n_vocab} does not match config num_symbols={num_symbols}")
            #-------------
            # commented out this code because this is not supposed to happen if you used the preprocess.py script
            # uncomment if you want to use the encoder from checkpoint + update num_symbols in the .json file manually
            #-------------
            #if ckpt.model_g.n_vocab > num_symbols and not discard_encoder:
            #    num_symbols = ckpt.model_g.n_vocab
            #    _LOGGER.info(f"Training with num_symbols={num_symbols}")
            ###############
        if ckpt.model_g.n_speakers != num_speakers:
            _LOGGER.warning(f"Checkpoint num_speakers={ckpt.model_g.n_speakers} does not match config num_speakers={num_speakers}")
            #num_speakers = ckpt.model_g.n_speakers
        if ckpt.hparams.sample_rate != sample_rate:
            _LOGGER.warning(f"Checkpoint sample_rate={ckpt.hparams.sample_rate} does not match config sample_rate={sample_rate}")
            #sample_rate = ckpt.hparams.sample_rate

    model = VitsModel(
        num_symbols=num_symbols,
        num_speakers=num_speakers,
        sample_rate=sample_rate,
        dataset=[dataset_path],
        **dict_args,
    )
    _LOGGER.info(f"VitsModel params: num_symbols={num_symbols} num_speakers={num_speakers} sample_rate={sample_rate}")

    if resume_from_checkpoint:
        saved_state_dict = ckpt.state_dict()

        # Filter the state dictionary by removing the encoder weights
        enc_key = 'model_g.enc_p.emb.weight'
        if enc_key in saved_state_dict:
            saved_shape = saved_state_dict[enc_key].shape
            current_shape = model.state_dict()[enc_key].shape
            if saved_shape[0] != current_shape[0]:
                _LOGGER.warning(
                    "Size mismatch detected for '%s': saved shape %s vs current shape %s. ",
                    enc_key, saved_shape, current_shape
                )
                discard_encoder = True

            if discard_encoder:
                _LOGGER.warning(
                    "Skipping encoder weights from the checkpoint. (will be randomly initialized)"
                )
                saved_state_dict.pop(enc_key)

        load_state_dict(model, saved_state_dict)
        _LOGGER.info("Successfully loaded model weights.")

    if resume_from_single_speaker_checkpoint:
        assert num_speakers > 1, "--resume-from-single-speaker-checkpoint is only for multi-speaker models."
        _LOGGER.info('Resuming from single-speaker checkpoint: %s', resume_from_single_speaker_checkpoint)

        model_single = VitsModel.load_from_checkpoint(resume_from_single_speaker_checkpoint, dataset=None)
        g_dict = model_single.model_g.state_dict()

        for key in list(g_dict.keys()):
            if key.startswith('dec.cond') or key.startswith('dp.cond') or ('enc.cond_layer' in key):
                g_dict.pop(key, None)

        load_state_dict(model.model_g, g_dict)
        load_state_dict(model.model_d, model_single.model_d.state_dict())
        _LOGGER.info('Successfully converted single-speaker checkpoint to multi-speaker')

    _LOGGER.info('training started!!')
    trainer.fit(model)


if __name__ == '__main__':
    main()





from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.thirdparty.hangul2ipa import hangul2ipa
from phoonnx.config import Alphabet


class G2PKPhonemizer(BasePhonemizer):

    def __init__(self, descriptive=True, group_vowels=True, to_syl=True,
                 alphabet=Alphabet.IPA):
        assert alphabet in [Alphabet.IPA, Alphabet.HANGUL]
        from g2pk import G2p
        self.g2p = G2p()
        self.descriptive = descriptive
        self.group_vowels = group_vowels
        self.to_syl = to_syl
        super().__init__(alphabet)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["ko"])

    def phonemize_string(self, text: str, lang: str = "ko") -> str:
        """
        """
        lang = self.get_lang(lang)
        p = self.g2p(text, descriptive=self.descriptive,
                     group_vowels=self.group_vowels,
                     to_syl=self.to_syl)
        if self.alphabet == Alphabet.IPA:
            return hangul2ipa(p)
        return p


class KoG2PPhonemizer(BasePhonemizer):
    """https://github.com/scarletcho/KoG2P"""
    def __init__(self,   alphabet=Alphabet.IPA):
        assert alphabet in [Alphabet.IPA, Alphabet.HANGUL]
        from phoonnx.thirdparty.kog2p import runKoG2P
        self.g2p = runKoG2P
        super().__init__(alphabet)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["ko"])

    def phonemize_string(self, text: str, lang: str = "ko") -> str:
        """
        """
        lang = self.get_lang(lang)
        p = self.g2p(text)
        if self.alphabet == Alphabet.IPA:
            return hangul2ipa(p)
        return p


if __name__ == "__main__":

    pho = G2PKPhonemizer(ipa=False)
    pho2 = KoG2PPhonemizer(ipa=False)
    lang = "ko"

    text = "터미널에서 원하는 문자열을 함께 입력해 사용할 수 있습니다."
    print(f"\n--- Getting phonemes for '{text}' ---")
    phonemes_cotovia = pho.phonemize(text, lang)
    print(f"  G2PK Phonemes: {phonemes_cotovia}")

    phonemes_cotovia = pho2.phonemize(text, lang)
    print(f"  KoG2P Phonemes: {phonemes_cotovia}")




#!/usr/bin/python
# -*- coding=utf-8 -*-
"""
Arabic module

Features:
=========
 - Arabic letters classification
 - Text tokenization
 - Strip Harakat (all, except Shadda, tatweel, last_haraka)
 - Sperate and  join Letters and Harakat
 - Reduce tashkeel
 - Mesure tashkeel similarity (Harakats, fully or partially vocalized, similarity with a template)
 - Letters normalization (Ligatures and Hamza)

@author: Taha Zerrouki
@contact: taha dot zerrouki at gmail dot com
@copyright: Arabtechies, Arabeyes, Taha Zerrouki
@license: GPL
@date:2010/03/01
@version: 0.1
"""
import re
import unicodedata

from . import stack

COMMA = "\u060c"
SEMICOLON = "\u061b"
QUESTION = "\u061f"
HAMZA = "\u0621"
ALEF_MADDA = "\u0622"
ALEF_HAMZA_ABOVE = "\u0623"
WAW_HAMZA = "\u0624"
ALEF_HAMZA_BELOW = "\u0625"
YEH_HAMZA = "\u0626"
ALEF = "\u0627"
BEH = "\u0628"
TEH_MARBUTA = "\u0629"
TEH = "\u062a"
THEH = "\u062b"
JEEM = "\u062c"
HAH = "\u062d"
KHAH = "\u062e"
DAL = "\u062f"
THAL = "\u0630"
REH = "\u0631"
ZAIN = "\u0632"
SEEN = "\u0633"
SHEEN = "\u0634"
SAD = "\u0635"
DAD = "\u0636"
TAH = "\u0637"
ZAH = "\u0638"
AIN = "\u0639"
GHAIN = "\u063a"
TATWEEL = "\u0640"
FEH = "\u0641"
QAF = "\u0642"
KAF = "\u0643"
LAM = "\u0644"
MEEM = "\u0645"
NOON = "\u0646"
HEH = "\u0647"
WAW = "\u0648"
ALEF_MAKSURA = "\u0649"
YEH = "\u064a"
MADDA_ABOVE = "\u0653"
HAMZA_ABOVE = "\u0654"
HAMZA_BELOW = "\u0655"
ZERO = "\u0660"
ONE = "\u0661"
TWO = "\u0662"
THREE = "\u0663"
FOUR = "\u0664"
FIVE = "\u0665"
SIX = "\u0666"
SEVEN = "\u0667"
EIGHT = "\u0668"
NINE = "\u0669"
ZERO_W = "\u0030"
ONE_W = "\u0031"
TWO_W = "\u0032"
THREE_W = "\u0033"
FOUR_W = "\u0034"
FIVE_W = "\u0035"
SIX_W = "\u0036"
SEVEN_W = "\u0037"
EIGHT_W = "\u0038"
NINE_W = "\u0039"
ZERO_P = "\u06f0"
ONE_P = "\u06f1"
TWO_P = "\u06f2"
THREE_P = "\u06f3"
FOUR_P = "\u06f4"
FIVE_P = "\u06f5"
SIX_P = "\u06f6"
SEVEN_P = "\u06f7"
EIGHT_P = "\u06f8"
NINE_P = "\u06f9"
PERCENT = "\u066a"
DECIMAL = "\u066b"
THOUSANDS = "\u066c"
STAR = "\u066d"
MINI_ALEF = "\u0670"
ALEF_WASLA = "\u0671"
FULL_STOP = "\u06d4"
BYTE_ORDER_MARK = "\ufeff"

# Diacritics
FATHATAN = "\u064b"
DAMMATAN = "\u064c"
KASRATAN = "\u064d"
FATHA = "\u064e"
DAMMA = "\u064f"
KASRA = "\u0650"
SHADDA = "\u0651"
SUKUN = "\u0652"

# Small Letters
SMALL_ALEF = "\u0670"
SMALL_WAW = "\u06e5"
SMALL_YEH = "\u06e6"
# Ligatures
LAM_ALEF = "\ufefb"
LAM_ALEF_HAMZA_ABOVE = "\ufef7"
LAM_ALEF_HAMZA_BELOW = "\ufef9"
LAM_ALEF_MADDA_ABOVE = "\ufef5"
SIMPLE_LAM_ALEF = "\u0644\u0627"
SIMPLE_LAM_ALEF_HAMZA_ABOVE = "\u0644\u0623"
SIMPLE_LAM_ALEF_HAMZA_BELOW = "\u0644\u0625"
SIMPLE_LAM_ALEF_MADDA_ABOVE = "\u0644\u0622"
# groups
LETTERS = "".join(
    [
        ALEF,
        BEH,
        TEH,
        TEH_MARBUTA,
        THEH,
        JEEM,
        HAH,
        KHAH,
        DAL,
        THAL,
        REH,
        ZAIN,
        SEEN,
        SHEEN,
        SAD,
        DAD,
        TAH,
        ZAH,
        AIN,
        GHAIN,
        FEH,
        QAF,
        KAF,
        LAM,
        MEEM,
        NOON,
        HEH,
        WAW,
        ALEF_MAKSURA,
        YEH,
        HAMZA,
        ALEF_MADDA,
        ALEF_HAMZA_ABOVE,
        WAW_HAMZA,
        ALEF_HAMZA_BELOW,
        YEH_HAMZA,
    ]
)

NUMBERS_EAST = (ZERO, ONE, TWO, THREE, FOUR, FIVE, SIX, SEVEN, EIGHT, NINE)
NUMBERS_WEST = (
    ZERO_W,
    ONE_W,
    TWO_W,
    THREE_W,
    FOUR_W,
    FIVE_W,
    SIX_W,
    SEVEN_W,
    EIGHT_W,
    NINE_W,
)
NUMBERS_PERS = (
    ZERO_P,
    ONE_P,
    TWO_P,
    THREE_P,
    FOUR_P,
    FIVE_P,
    SIX_P,
    SEVEN_P,
    EIGHT_P,
    NINE_P,
)


TASHKEEL = (FATHATAN, DAMMATAN, KASRATAN, FATHA, DAMMA, KASRA, SUKUN, SHADDA)
HARAKAT = (FATHATAN, DAMMATAN, KASRATAN, FATHA, DAMMA, KASRA, SUKUN)

SHORTHARAKAT = (FATHA, DAMMA, KASRA, SUKUN)

TANWIN = (FATHATAN, DAMMATAN, KASRATAN)

NOT_DEF_HARAKA = TATWEEL

LIGUATURES = (
    LAM_ALEF,
    LAM_ALEF_HAMZA_ABOVE,
    LAM_ALEF_HAMZA_BELOW,
    LAM_ALEF_MADDA_ABOVE,
)

HAMZAT = (
    HAMZA,
    WAW_HAMZA,
    YEH_HAMZA,
    HAMZA_ABOVE,
    HAMZA_BELOW,
    ALEF_HAMZA_BELOW,
    ALEF_HAMZA_ABOVE,
)

ALEFAT = (
    ALEF,
    ALEF_MADDA,
    ALEF_HAMZA_ABOVE,
    ALEF_HAMZA_BELOW,
    ALEF_WASLA,
    ALEF_MAKSURA,
    SMALL_ALEF,
)

WEAK = (ALEF, WAW, YEH, ALEF_MAKSURA)
YEHLIKE = (YEH, YEH_HAMZA, ALEF_MAKSURA, SMALL_YEH)

WAWLIKE = (WAW, WAW_HAMZA, SMALL_WAW)
TEHLIKE = (TEH, TEH_MARBUTA)

SMALL = (SMALL_ALEF, SMALL_WAW, SMALL_YEH)

MOON = (
    HAMZA,
    ALEF_MADDA,
    ALEF_HAMZA_ABOVE,
    ALEF_HAMZA_BELOW,
    ALEF,
    BEH,
    JEEM,
    HAH,
    KHAH,
    AIN,
    GHAIN,
    FEH,
    QAF,
    KAF,
    MEEM,
    HEH,
    WAW,
    YEH,
)

SUN = (
    TEH,
    THEH,
    DAL,
    THAL,
    REH,
    ZAIN,
    SEEN,
    SHEEN,
    SAD,
    DAD,
    TAH,
    ZAH,
    LAM,
    NOON,
)

ALPHABETIC_ORDER = {
    ALEF: 1,
    BEH: 2,
    TEH: 3,
    TEH_MARBUTA: 3,
    THEH: 4,
    JEEM: 5,
    HAH: 6,
    KHAH: 7,
    DAL: 8,
    THAL: 9,
    REH: 10,
    ZAIN: 11,
    SEEN: 12,
    SHEEN: 13,
    SAD: 14,
    DAD: 15,
    TAH: 16,
    ZAH: 17,
    AIN: 18,
    GHAIN: 19,
    FEH: 20,
    QAF: 21,
    KAF: 22,
    LAM: 23,
    MEEM: 24,
    NOON: 25,
    HEH: 26,
    WAW: 27,
    YEH: 28,
    HAMZA: 29,
    ALEF_MADDA: 29,
    ALEF_HAMZA_ABOVE: 29,
    WAW_HAMZA: 29,
    ALEF_HAMZA_BELOW: 29,
    YEH_HAMZA: 29,
}

NAMES = {
    ALEF: "ألف",
    BEH: "باء",
    TEH: "تاء",
    TEH_MARBUTA: "تاء مربوطة",
    THEH: "ثاء",
    JEEM: "جيم",
    HAH: "حاء",
    KHAH: "خاء",
    DAL: "دال",
    THAL: "ذال",
    REH: "راء",
    ZAIN: "زاي",
    SEEN: "سين",
    SHEEN: "شين",
    SAD: "صاد",
    DAD: "ضاد",
    TAH: "طاء",
    ZAH: "ظاء",
    AIN: "عين",
    GHAIN: "غين",
    FEH: "فاء",
    QAF: "قاف",
    KAF: "كاف",
    LAM: "لام",
    MEEM: "ميم",
    NOON: "نون",
    HEH: "هاء",
    WAW: "واو",
    YEH: "ياء",
    HAMZA: "همزة",
    TATWEEL: "تطويل",
    ALEF_MADDA: "ألف ممدودة",
    ALEF_MAKSURA: "ألف مقصورة",
    ALEF_HAMZA_ABOVE: "همزة على الألف",
    WAW_HAMZA: "همزة على الواو",
    ALEF_HAMZA_BELOW: "همزة تحت الألف",
    YEH_HAMZA: "همزة على الياء",
    FATHATAN: "فتحتان",
    DAMMATAN: "ضمتان",
    KASRATAN: "كسرتان",
    FATHA: "فتحة",
    DAMMA: "ضمة",
    KASRA: "كسرة",
    SHADDA: "شدة",
    SUKUN: "سكون",
}
HAMZAT_STRING = "".join(HAMZAT)
HARAKAT_STRING = "".join(HARAKAT)
TASHKEEL_STRING = "".join(TASHKEEL)
# regular expretion

HARAKAT_PATTERN = re.compile("[" + "".join(HARAKAT) + "]", re.UNICODE)
# ~ """ pattern to strip Harakat"""
LASTHARAKA_PATTERN = re.compile(
    "[%s]$|[%s]" % ("".join(HARAKAT), "".join(TANWIN)), re.UNICODE
)
# ~ """ Pattern to strip only the last haraka """
SHORTHARAKAT_PATTERN = re.compile("[" + "".join(SHORTHARAKAT) + "]", re.UNICODE)
# ~ Pattern to lookup Short Harakat(Fatha, Damma, Kasra, sukun, tanwin),
# but not shadda
TASHKEEL_PATTERN = re.compile("[" + "".join(TASHKEEL) + "]", re.UNICODE)
# ~ """ Harakat and shadda pattern  """
HAMZAT_PATTERN = re.compile("[" + "".join(HAMZAT) + "]", re.UNICODE)
# ~ """ all hamzat pattern"""
ALEFAT_PATTERN = re.compile("[" + "".join(ALEFAT) + "]", re.UNICODE)
# ~ """ all alef like letters """
LIGUATURES_PATTERN = re.compile("[" + "".join(LIGUATURES) + "]", re.UNICODE)
# ~ """ all liguatures pattern """
TOKEN_PATTERN = re.compile(r"([^\w\u0670\u064b-\u0652']+)", re.UNICODE)
# ~ """ pattern to tokenize a text"""
TOKEN_PATTERN_SPLIT = re.compile(r"([\w\u0670\u064b-\u0652']+)", re.UNICODE)
# ~ """ pattern to tokenize a text with positions"""
TOKEN_REPLACE = re.compile("\t|\r|\f|\v| ")

# Arabic string
ARABIC_STRING = re.compile(
    r"([^\u0600-\u0652%s%s%s\s\d])"
    % (LAM_ALEF, LAM_ALEF_HAMZA_ABOVE, LAM_ALEF_MADDA_ABOVE),
    re.UNICODE,
)
# Arabic range
ARABIC_RANGE = re.compile(
    "([^\u0600-\u06ff\ufb50-\ufdff\ufe70-\ufeff\u0750-\u077f])", re.UNICODE
)

# Space fixes
FIX_SPACES_PAT = re.compile(r"\s*([?؟!.,،:]+(?:\s+[?؟!.,،:]+)*)\s*", re.UNICODE)

DIACRITICS = [
    chr(x) for x in range(0x0600, 0x06FF) if unicodedata.category(chr(x)) == "Mn"
]
# ~ \u0610   ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM
# ~ \u0611   ARABIC SIGN ALAYHE ASSALLAM
# ~ \u0612   ARABIC SIGN RAHMATULLAH ALAYHE
# ~ \u0613   ARABIC SIGN RADI ALLAHOU ANHU
# ~ \u0614   ARABIC SIGN TAKHALLUS
# ~ \u0615   ARABIC SMALL HIGH TAH
# ~ \u0616   ARABIC SMALL HIGH LIGATURE ALEF WITH LAM WITH YEH
# ~ \u0617   ARABIC SMALL HIGH ZAIN
# ~ \u0618   ARABIC SMALL FATHA
# ~ \u0619   ARABIC SMALL DAMMA
# ~ \u061a   ARABIC SMALL KASRA
# ~ \u064b   ARABIC FATHATAN
# ~ \u064c   ARABIC DAMMATAN
# ~ \u064d   ARABIC KASRATAN
# ~ \u064e   ARABIC FATHA
# ~ \u064f   ARABIC DAMMA
# ~ \u0650   ARABIC KASRA
# ~ \u0651   ARABIC SHADDA
# ~ \u0652   ARABIC SUKUN
# ~ \u0653   ARABIC MADDAH ABOVE
# ~ \u0654   ARABIC HAMZA ABOVE
# ~ \u0655   ARABIC HAMZA BELOW
# ~ \u0656   ARABIC SUBSCRIPT ALEF
# ~ \u0657   ARABIC INVERTED DAMMA
# ~ \u0658   ARABIC MARK NOON GHUNNA
# ~ \u0659   ARABIC ZWARAKAY
# ~ \u065a   ARABIC VOWEL SIGN SMALL V ABOVE
# ~ \u065b   ARABIC VOWEL SIGN INVERTED SMALL V ABOVE
# ~ \u065c   ARABIC VOWEL SIGN DOT BELOW
# ~ \u065d   ARABIC REVERSED DAMMA
# ~ \u065e   ARABIC FATHA WITH TWO DOTS
# ~ \u0670   ARABIC LETTER SUPERSCRIPT ALEF
# ~ \u06d6   ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA
# ~ \u06d7   ARABIC SMALL HIGH LIGATURE QAF WITH LAM WITH ALEF MAKSURA
# ~ \u06d8   ARABIC SMALL HIGH MEEM INITIAL FORM
# ~ \u06d9   ARABIC SMALL HIGH LAM ALEF
# ~ \u06da   ARABIC SMALL HIGH JEEM
# ~ \u06db   ARABIC SMALL HIGH THREE DOTS
# ~ \u06dc   ARABIC SMALL HIGH SEEN
# ~ \u06df   ARABIC SMALL HIGH ROUNDED ZERO
# ~ \u06e0   ARABIC SMALL HIGH UPRIGHT RECTANGULAR ZERO
# ~ \u06e1   ARABIC SMALL HIGH DOTLESS HEAD OF KHAH
# ~ \u06e2   ARABIC SMALL HIGH MEEM ISOLATED FORM
# ~ \u06e3   ARABIC SMALL LOW SEEN
# ~ \u06e4   ARABIC SMALL HIGH MADDA
# ~ \u06e7   ARABIC SMALL HIGH YEH
# ~ \u06e8   ARABIC SMALL HIGH NOON
# ~ \u06ea   ARABIC EMPTY CENTRE LOW STOP
# ~ \u06eb   ARABIC EMPTY CENTRE HIGH STOP
# ~ \u06ec   ARABIC ROUNDED HIGH STOP WITH FILLED CENTRE
# ~ \u06ed   ARABIC SMALL LOW MEEM


################################################
# { is letter functions
################################################
def is_sukun(archar):
    """Checks if the given ``archar``Sukun Mark."""
    return archar == SUKUN


def is_shadda(archar):
    """Checks if the given ``archar`` is  Shadda Mark."""
    return archar == SHADDA


def is_tatweel(archar):
    """Checks if the given ``archar`` Tatweel letter modifier."""
    return archar == TATWEEL


def is_tanwin(archar):
    """Checks if the given ``archar`` Tanwin Marks"""
    return archar in TANWIN


def is_tashkeel(archar):
    """Checks if the given ``archar`` Arabic Tashkeel Marks (
    - FATHA, DAMMA, KASRA, SUKUN,
    - SHADDA,
    - FATHATAN, DAMMATAN, KASRATAn)."""
    return archar in TASHKEEL


def is_haraka(archar):
    """Checks if the given ``archar`` Arabic Harakat Marks (FATHA, DAMMA, KASRA, SUKUN, TANWIN)."""
    return archar in HARAKAT


def is_shortharaka(archar):
    """Checks if the given ``archar``  short Harakat Marks (FATHA, DAMMA, KASRA, SUKUN)."""
    return archar in SHORTHARAKAT


def is_ligature(archar):
    """Checks for Arabic  Ligatures like LamAlef.
    (LAM_ALEF, LAM_ALEF_HAMZA_ABOVE, LAM_ALEF_HAMZA_BELOW, LAM_ALEF_MADDA_ABOVE)
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in LIGUATURES


def is_hamza(archar):
    """Checks for Arabic  Hamza forms.
    HAMZAT are (HAMZA, WAW_HAMZA, YEH_HAMZA, HAMZA_ABOVE, HAMZA_BELOW,
    ALEF_HAMZA_BELOW, ALEF_HAMZA_ABOVE)
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in HAMZAT


def is_alef(archar):
    """Checks for Arabic Alef forms.
    ALEFAT = (ALEF, ALEF_MADDA, ALEF_HAMZA_ABOVE, ALEF_HAMZA_BELOW, ALEF_WASLA, ALEF_MAKSURA)
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in ALEFAT


def is_yehlike(archar):
    """Checks for Arabic Yeh forms.
    Yeh forms : YEH, YEH_HAMZA, SMALL_YEH, ALEF_MAKSURA
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in YEHLIKE


def is_wawlike(archar):
    """Checks for Arabic Waw like forms.
    Waw forms : WAW, WAW_HAMZA, SMALL_WAW
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in WAWLIKE


def is_teh(archar):
    """Checks for Arabic Teh forms.
    Teh forms : TEH, TEH_MARBUTA
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in TEHLIKE


def is_small(archar):
    """Checks for Arabic Small letters.
    SMALL Letters : SMALL ALEF, SMALL WAW, SMALL YEH
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in SMALL


def is_weak(archar):
    """Checks for Arabic Weak letters.
    Weak Letters : ALEF, WAW, YEH, ALEF_MAKSURA
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in WEAK


def is_moon(archar):
    """Checks for Arabic Moon letters.
    Moon Letters :
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in MOON


def is_sun(archar):
    """Checks for Arabic Sun letters.
    Moon Letters :
    @param archar: arabic unicode char
    @type archar: unicode
    @return:
    @rtype:Boolean
    """
    return archar in SUN


#####################################
# { general  letter functions
#####################################
def order(archar):
    """return Arabic letter order between 1 and 29.
    Alef order is 1, Yeh is 28, Hamza is 29.
    Teh Marbuta has the same ordre with Teh, 3.
    @param archar: arabic unicode char
    @type archar: unicode
    @return: arabic order.
    @rtype: integer
    """
    return ALPHABETIC_ORDER.get(archar, 0)


def name(archar, default_name=""):
    """return Arabic letter name in arabic.     Alef order is 1, Yeh is 28,
    Hamza is 29. Teh Marbuta has the same ordre with Teh, 3.
    @param archar: arabic unicode char
    @type archar: unicode
    @param default_name: arabic unicode char
    @type default_name: unicode
    @return: arabic name.
    @rtype: unicode
    """
    return NAMES.get(archar, default_name)


def arabicrange():
    """return a list of arabic characteres .
    Return a list of characteres between \u060c to \u0652
    @return: list of arabic characteres.
    @rtype: unicode
    """
    mylist = []
    for i in range(0x0600, 0x00653):
        try:
            mylist.append(chr(i))
        except NameError:
            # python 3 compatible
            mylist.append(chr(i))
        except ValueError:
            pass
    return mylist


#####################################
# { Has letter functions
#####################################
def has_shadda(word):
    """Checks if the arabic word  contains shadda.
    @param word: arabic unicode char
    @type word: unicode
    @return: if shadda exists
    @rtype:Boolean
    """
    if re.search(SHADDA, word):
        return True
    return False


#####################################
# { word and text functions
#####################################
def is_vocalized(word):
    """Checks if the arabic word is vocalized.
    the word musn't  have any spaces and pounctuations.
    @param word: arabic unicode char
    @type word: unicode
    @return: if the word is vocalized
    @rtype:Boolean
    """
    if word.isalpha():
        return False
    for char in word:
        if is_tashkeel(char):
            break
    else:
        return False
    return True


def is_vocalizedtext(text):
    """Checks if the arabic text is vocalized.
    The text can contain many words and spaces
    @param text: arabic unicode char
    @type text: unicode
    @return: if the word is vocalized
    @rtype:Boolean
    """
    return bool(re.search(HARAKAT_PATTERN, text))


def is_arabicstring(text):
    """Checks for an  Arabic standard Unicode block characters
    An arabic string can contain spaces, digits and pounctuation.
    but only arabic standard characters, not extended arabic
    @param text: input text
    @type text: unicode
    @return: True if all charaters are in Arabic block
    @rtype: Boolean
    """
    if ARABIC_STRING.search(text):
        return False
    return True


def is_arabicrange(text):
    """Checks for an  Arabic Unicode block characters
    @param text: input text
    @type text: unicode
    @return: True if all charaters are in Arabic block
    @rtype: Boolean
    """
    if ARABIC_RANGE.search(text):
        return False
    return True


def is_arabicword(word):
    """Checks for an valid Arabic  word.
    An Arabic word not contains spaces, digits and pounctuation
    avoid some spelling error, TEH_MARBUTA must be at the end.
    @param word: input word
    @type word: unicode
    @return: True if all charaters are in Arabic block
    @rtype: Boolean
    """
    if len(word) == 0:
        return False
    elif re.search(
        "([^\u0600-\u0652%s%s%s])"
        % (LAM_ALEF, LAM_ALEF_HAMZA_ABOVE, LAM_ALEF_MADDA_ABOVE),
        word,
    ):
        return False
    elif is_haraka(word[0]) or word[0] in (WAW_HAMZA, YEH_HAMZA):
        return False
    #  if Teh Marbuta or Alef_Maksura not in the end
    elif re.match("^(.)*[%s](.)+$" % ALEF_MAKSURA, word):
        return False
    elif re.match(
        "^(.)*[%s]([^%s%s%s])(.)+$" % (TEH_MARBUTA, DAMMA, KASRA, FATHA), word
    ):
        return False
    elif re.search(SHADDA + SHADDA, word):
        return False
    else:
        return True


#####################################
# {Char functions
#####################################
def first_char(word):
    """
    Return the first char
    @param word: given word
    @type word: unicode
    @return: the first char
    @rtype: unicode char
    """
    return word[0]


def second_char(word):
    """
    Return the second char
    @param word: given word
    @type word: unicode
    @return: the first char
    @rtype: unicode char
    """
    return word[1:2]


def last_char(word):
    """
    Return the last letter
    example: zerrouki; 'i' is the last.
    @param word: given word
    @type word: unicode
    @return: the last letter
    @rtype: unicode char
    """
    return word[-1:]


def secondlast_char(word):
    """
    Return the second last letter example: zerrouki; 'k' is the second last.
    @param word: given word
    @type word: unicode
    @return: the second last letter
    @rtype: unicode char
    """
    return word[-2:-1]


#####################################
# {Strip functions
#####################################
def strip_harakat(text):
    """Strip Harakat from arabic word except Shadda.
    The striped marks are :
        - FATHA, DAMMA, KASRA
        - SUKUN
        - FATHATAN, DAMMATAN, KASRATAN,

    Example:
        >>> text = u"الْعَرَبِيّةُ"
        >>> strip_harakat(text)
        >>> العربيّة

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.
    """
    # if text:
    # return  re.sub(HARAKAT_PATTERN, u'', text)
    # return text
    if not text:
        return text
    elif is_vocalized(text):
        for char in HARAKAT:
            text = text.replace(char, "")
    return text


def strip_lastharaka(text):
    """Strip the last Haraka from arabic word except Shadda.
    The striped marks are :
        - FATHA, DAMMA, KASRA
        - SUKUN
        - FATHATAN, DAMMATAN, KASRATAN

    Example:
        >>> text = u"الْعَرَبِيّةُ"
        >>> strip_lastharaka(text)
        الْعَرَبِيّة

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.
    """
    if text:
        if is_vocalized(text):
            return re.sub(LASTHARAKA_PATTERN, "", text)
    return text


def strip_tashkeel(text):
    """Strip vowels from a text, include Shadda.
    The striped marks are :
        - FATHA, DAMMA, KASRA
        - SUKUN
        - SHADDA
        - FATHATAN, DAMMATAN, KASRATAN,, , .

    Example:
        >>> text = u"الْعَرَبِيّةُ"
        >>> strip_tashkeel(text)
        العربية

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.
    """
    if not text:
        return text
    elif is_vocalized(text):
        for char in TASHKEEL:
            text = text.replace(char, "")
    return text


def strip_small(text):
    """Strip small_letters from a text
    The striped marks are :
        - Small Alef الألف الخنجرية, .
        -Small WAW
        -Small Yeh
    Example:
        >>> text = u"الرحمن\u0670"
        >>> strip_small(text)
        الرحمن

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.
    """
    if not text:
        return text
    for char in SMALL:
        text = text.replace(char, "")
    return text


def strip_tatweel(text):
    """
    Strip tatweel from a text and return a result text.

    Example:
        >>> text = u"العـــــربية"
        >>> strip_tatweel(text)
        العربية

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.

    """
    return text.replace(TATWEEL, "")


def strip_shadda(text):
    """
    Strip Shadda from a text and return a result text.

    Example:
        >>> text = u"الشّمسيّة"
        >>> strip_shadda(text)
         الشمسية

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.
    """
    return text.replace(SHADDA, "")


def strip_diacritics(text):
    """Strip arabic diacritics from a text
    The striped marks are :
        - Small Alef الألف الخنجرية, .
        - Harakat + Shadda
        - Quranic marks
        - Extended arabic diacritics
    Example:
        >>> text = u"الرحمن\u0670"
        >>> strip_small(text)
        الرحمن

    @param text: arabic text.
    @type text: unicode.
    @return: return a striped text.
    @rtype: unicode.
    """
    if not text:
        return text
    for char in DIACRITICS:
        text = text.replace(char, "")
    return text


def normalize_ligature(text):
    """Normalize Lam Alef ligatures into two letters (LAM and ALEF),
    and Tand return a result text.
    Some systems present lamAlef ligature as a single letter,
    this function convert it into two letters,
    The converted letters into  LAM and ALEF are :
        - LAM_ALEF, LAM_ALEF_HAMZA_ABOVE, LAM_ALEF_HAMZA_BELOW, LAM_ALEF_MADDA_ABOVE

    Example:
        >>> text = u"لانها لالء الاسلام"
        >>> normalize_ligature(text)
        لانها لالئ الاسلام

    @param text: arabic text.
    @type text: unicode.
    @return: return a converted text.
    @rtype: unicode.
    """
    if text:
        return LIGUATURES_PATTERN.sub("%s%s" % (LAM, ALEF), text)
    return text


def normalize_hamza(word, method="uniform"):
    """Standardize the Hamzat into one form of hamza,
    replace Madda by hamza and alef.
    Replace the LamAlefs by simplified letters.


    Example:
        >>> import pyarabic.araby as araby
        >>> text1 = u"جاء سؤال الأئمة عن الإسلام آجلا"
        >>> araby.normalize_hamza(text1)
        'جاء سءال الءءمة عن الءسلام ءءجلا'
        >>> araby.normalize_hamza(text1, method="tasheel")
            'جاء سوال الايمة عن الاسلام ا
    @param word: arabic text.
    @type word: unicode.
    @param method: how to convert hamzat (uniform, tasheel).
    @type method: unicode.
    @return: return a converted text.
    @rtype: unicode.
    """
    if method == "tasheel" or method == "تسهيل":
        # Alefat to Alef
        word = word.replace(ALEF_MADDA, ALEF)
        word = word.replace(ALEF_HAMZA_ABOVE, ALEF)
        word = word.replace(ALEF_HAMZA_BELOW, ALEF)
        word = word.replace(HAMZA_ABOVE, ALEF)
        word = word.replace(HAMZA_BELOW, ALEF)
        # on Waw
        word = word.replace(WAW_HAMZA, WAW)
        # on Yeh
        word = word.replace(YEH_HAMZA, YEH)
    else:
        if word.startswith(ALEF_MADDA):
            if (
                len(word) >= 3
                and (word[1] not in HARAKAT)
                and (word[2] == SHADDA or len(word) == 3)
            ):
                word = HAMZA + ALEF + word[1:]
            else:
                word = HAMZA + HAMZA + word[1:]
        # convert all Hamza from into one form
        word = word.replace(ALEF_MADDA, HAMZA + HAMZA)
        word = HAMZAT_PATTERN.sub(HAMZA, word)
    return word


def normalize_teh(text):
    """
    converts TEH_MARBUTA to HEH
    Example:
        >>> text = 'محبة'
        >>> normalize_teh(text)
        'محبه'

    """
    return re.sub("[" + "".join(TEH_MARBUTA) + "]", HEH, text)


def normalize_alef(text):
    """
    converts all alefs to ALEF_MAMDODA with the exception of Alef maksura

    """
    # a small alef before Alef Maksura is ommited
    text = text.replace(SMALL_ALEF + ALEF_MAKSURA, ALEF_MAKSURA)
    text = text.replace(ALEF_MAKSURA + SMALL_ALEF, ALEF_MAKSURA)
    return re.sub(ALEFAT_PATTERN, ALEF, text)


def separate(word, extract_shadda=False):
    """
    separate the letters from the vowels, in arabic word,
    if a letter hasn't a haraka, the not definited haraka is attributed.
    return (letters, vowels)

    Example:
        >>> araby.separate(text)
        (u'\u0627\u0644\u0639\u0631\u0628\u064a\u0629',
        u'\u064e\u0652\u064e\u064e\u064e\u064e\u064f')
        >>> letters, marks =araby.separate(text)
        >>> print letters.encode('utf8')
        العربية
        >>> print marks.encode('utf8')
        >>> for m in marks:
        ...     print araby.name(m)
        فتحة
        سكون
        فتحة
        فتحة
        فتحة
        فتحة
        ضمة

    @param word: the input word
    @type word: unicode
    @param extract_shadda: extract shadda as seperate text
    @type extract_shadda: Boolean
    @return: (letters, vowels)
    @rtype:couple of unicode
    """
    stack1 = stack.Stack(word)
    # the word is inversed in the stack
    stack1.items.reverse()
    letters = stack.Stack()
    marks = stack.Stack()
    vowels = HARAKAT
    last1 = stack1.pop()
    # if the last element must be a letter,
    # the arabic word can't starts with a haraka
    # in th stack the word is inversed
    while last1 in vowels:
        last1 = stack1.pop()
    while last1 != None:
        if last1 in vowels:
            # we can't have two harakats beside.
            # the shadda is considered as a letter
            marks.pop()
            marks.push(last1)
        elif last1 == SHADDA:
            # is the element is a Shadda,
            # the previous letter must have a sukun as mark,
            # and the shadda take the indefinate  mark
            marks.pop()
            marks.push(SUKUN)
            marks.push(NOT_DEF_HARAKA)
            letters.push(SHADDA)
        else:
            marks.push(NOT_DEF_HARAKA)
            letters.push(last1)
        last1 = stack1.pop()
    if extract_shadda:
        # the shadda is considered as letter
        wordletters = "".join(letters.items)
        # print wordletters.encode('utf8')
        shaddaplaces = re.sub("[^%s]" % SHADDA, TATWEEL, wordletters)
        shaddaplaces = re.sub("%s%s" % (TATWEEL, SHADDA), SHADDA, shaddaplaces)
        # print wordletters.encode('utf8')
        wordletters = strip_shadda(wordletters)
        # print wordletters.encode('utf8')
        return (wordletters, "".join(marks.items), shaddaplaces)
    else:
        return ("".join(letters.items), "".join(marks.items))


def joint(letters, marks):
    """joint the letters with the marks
    the length ot letters and marks must be equal
    return word

    Example:
        >>> letters = u"العربية"
        >>> marks = u'\u064e\u0652\u064e\u064e\u064e\u064e\u064f'
        >>> word = araby.joint(letters, marks)
        >>> print word.encode('utf8')
        اَلْعَرَبَيَةُ

    @param letters: the word letters
    @type letters: unicode
    @param marks: the word marks
    @type marks: unicode
    @return: word
    @rtype: unicode
    """
    # The length ot letters and marks must be equal
    if len(letters) != len(marks):
        return ""
    stack_letter = stack.Stack(letters)
    stack_letter.items.reverse()
    stack_mark = stack.Stack(marks)
    stack_mark.items.reverse()

    word_stack = stack.Stack()
    last_letter = stack_letter.pop()
    last_mark = stack_mark.pop()
    vowels = HARAKAT
    while last_letter != None and last_mark != None:
        if last_letter == SHADDA:
            top = word_stack.pop()
            if top not in vowels:
                word_stack.push(top)
            word_stack.push(last_letter)
            if last_mark != NOT_DEF_HARAKA:
                word_stack.push(last_mark)
        else:
            word_stack.push(last_letter)
            if last_mark != NOT_DEF_HARAKA:
                word_stack.push(last_mark)

        last_letter = stack_letter.pop()
        last_mark = stack_mark.pop()

    if not (stack_letter.is_empty() and stack_mark.is_empty()):
        return False
    else:
        return "".join(word_stack.items)


def vocalizedlike(word1, word2):
    """
    if the two words has the same letters and the same harakats, this fuction return True.
    The two words can be full vocalized, or partial vocalized

    Example:
        >>> word1 = u"ضَربٌ"
        >>> word2 = u"ضَرْبٌ"
        >>> araby.vocalizedlike(word1, word2)
        True

    @param word1: first word
    @type word1: unicode
    @param word2: second word
    @type word2: unicode
    @return: if two words have similar vocalization
    @rtype: Boolean
    """
    if vocalized_similarity(word1, word2) < 0:
        return False
    else:
        return True


# -------------------------
# Function def vaznlike(word1, wazn):
# -------------------------
def waznlike(word1, wazn, extract_root=False):
    """If the  word1 is like a wazn (pattern), and can return root
    the letters must be equal,
    the wazn has FEH, AIN, LAM letters.
    this are as generic letters.
    The two words can be full vocalized, or partial vocalized

    Example:
        >>> word1 = u"ضارب"
        >>> wazn = u"فَاعِل"
        >>> araby.waznlike(word1, wazn)
        True

    @param word1: input word
    @type word1: unicode
    @param wazn: given word template  وزن
    @type wazn: unicode
    @param extract_root: return root if True
    @type extract_root: boolean
    @return: if two words have similar vocalization
    @rtype: Boolean
    """
    stack1 = stack.Stack(word1)
    stack2 = stack.Stack(wazn)
    root = stack.Stack()
    last1 = stack1.pop()
    last2 = stack2.pop()
    vowels = HARAKAT
    while last1 != None and last2 != None:
        if last1 == last2 and last2 not in (FEH, AIN, LAM):
            last1 = stack1.pop()
            last2 = stack2.pop()
        elif last1 not in vowels and last2 in (FEH, AIN, LAM):
            root.push(last1)
            # ~ print "t"
            last1 = stack1.pop()
            last2 = stack2.pop()
        elif last1 in vowels and last2 not in vowels:
            last1 = stack1.pop()
        elif last1 not in vowels and last2 in vowels:
            last2 = stack2.pop()
        else:
            break
    # reverse the root letters
    root.items.reverse()
    # ~ print " the root is ", root.items#"".join(root.items)
    if not (stack1.is_empty() and stack2.is_empty()):
        return False
    # if one letter is remind after pop in one stack
    elif last1 != None or last2 != None:
        return False
    else:
        if extract_root:
            return "".join(root.items)
        else:
            return True


def shaddalike(partial, fully):
    """
    If the two words has the same letters and the same harakats, this fuction return True.
    The first word is partially vocalized, the second is fully
    if the partially contians a shadda, it must be at the same place in the fully

    Example:
        >>> word1 = u"ردّ"
        >>> word2=u"ردَّ"
        >>> araby.shaddalike(word1, word2)
        True

    @param partial: the partially vocalized word
    @type partial: unicode
    @param fully: the fully vocalized word
    @type fully: unicode
    @return: if contains shadda
    @rtype: Boolean
    """
    # المدخل ليس به شدة، لا داعي للبحث
    if not has_shadda(partial):
        return True
    # المدخل به شدة، والنتيجة ليس بها شدة، خاطئ
    elif not has_shadda(fully) and has_shadda(partial):
        return False

    # المدخل والمخرج بهما شدة، نتأكد من موقعهما
    partial = strip_harakat(partial)
    fully = strip_harakat(fully)
    pstack = stack.Stack(partial)
    vstack = stack.Stack(fully)
    plast = pstack.pop()
    vlast = vstack.pop()
    # if debug: print "+0", Pstack, Vstack
    while plast != None and vlast != None:
        if plast == vlast:
            plast = pstack.pop()
            vlast = vstack.pop()
        elif plast == SHADDA and vlast != SHADDA:
            # if debug: print "+2", Pstack.items, Plast, Vstack.items, Vlast
            break
        elif plast != SHADDA and vlast == SHADDA:
            # if debug: print "+2", Pstack.items, Plast, Vstack.items, Vlast
            vlast = vstack.pop()
        else:
            # if debug: print "+2", Pstack.items, Plast, Vstack.items, Vlast
            break
    if not (pstack.is_empty() and vstack.is_empty()):
        return False
    else:
        return True


def reduce_tashkeel(text):
    """Reduce the Tashkeel, by deleting evident cases.

    Exmaple:
        >>> word = u"يُتَسََلَّمْنَ"
        >>> reduced = araby.reduce_tashkeel(word)
        >>> print reduced.encode('utf8')
        يُتسلّمن

    @param text: the input text fully vocalized.
    @type text: unicode.
    @return : partially vocalized text.
    @rtype: unicode.

    """
    patterns = [
        # delete all fathat, except on waw and yeh
        "(?<!(%s|%s))(%s|%s)"
        % (WAW, YEH, SUKUN, FATHA),  # delete damma if followed by waw.
        "%s(?=%s)" % (DAMMA, WAW),  # delete kasra if followed by yeh.
        "%s(?=%s)"
        % (KASRA, YEH),  # delete fatha if followed by alef to reduce yeh maftouha
        #  and waw maftouha before alef.
        "%s(?=%s)"
        % (
            FATHA,
            ALEF,
        ),  # delete fatha from yeh and waw if they are in the word begining.
        r"(?<=\s(%s|%s))%s"
        % (WAW, YEH, FATHA),  # delete kasra if preceded by Hamza below alef.
        "(?<=%s)%s" % (ALEF_HAMZA_BELOW, KASRA),
    ]
    reduced = text
    for pat in patterns:
        reduced = re.sub(pat, "", reduced)
    return reduced


def vocalized_similarity(word1, word2):
    """if the two words has the same letters and the same harakats, this function return True.
    The two words can be full vocalized, or partial vocalized

    Example:
        >>> word1 = u"ضَربٌ"
        >>> word2 = u"ضَرْبٌ"
        >>> araby.vocalizedlike(word1, word2)
        True
        >>> word1 = u"ضَربٌ"
        >>> word2 = u"ضَرْبٍ"
        >>> araby.vocalized_similarity(word1, word2)
        -1

    @param word1: first word
    @type word1: unicode
    @param word2: second word
    @type word2: unicode
    @return: return if words are similar, else return negative number of errors
    @rtype: Boolean / int
    """
    stack1 = stack.Stack(word1)
    stack2 = stack.Stack(word2)
    last1 = stack1.pop()
    last2 = stack2.pop()
    err_count = 0
    vowels = HARAKAT
    while last1 != None and last2 != None:
        if last1 == last2:
            last1 = stack1.pop()
            last2 = stack2.pop()
        elif last1 in vowels and last2 not in vowels:
            last1 = stack1.pop()
        elif last1 not in vowels and last2 in vowels:
            last2 = stack2.pop()
        else:
            # break
            if last1 == SHADDA:
                last1 = stack1.pop()
            elif last2 == SHADDA:
                last2 = stack2.pop()
            else:
                last1 = stack1.pop()
                last2 = stack2.pop()
                err_count += 1
    if err_count > 0:
        return -err_count
    else:
        return True


def sentence_tokenize(text):
    """
    Tokenize text into sentences.

    Example:
        >>> text = u"العربية لغة جميلة. والبلاد بعيدة، والشوق زائد"
        >>> tokens = araby.sentence_tokenize(text)
        >>> print(tokens)
        ‎‎['العربية لغة جميلة.', 'والبلاد بعيدة،', 'والشوق زائد']

    @param text: the input text.
    @type text: unicode.
    @return: list of sentences.
    @rtype: list.
    """
    text = re.sub("([.,:;،؟?\n])+([\n\t\r ])+", r"\1<SPLIT>", text, re.UNICODE)
    sentences = re.split("<SPLIT>", text)
    return sentences


def tokenize(text="", conditions=[], morphs=[]):
    """
    Tokenize text into words.

    Example:
        >>> text = u"العربية لغة جميلة."
        >>> tokens = araby.tokenize(text)
        >>> print u"\\n".join(tokens)
        ‎العربية
        ‎لغة
        ‎جميلة
        .

    Example 2 (To remove tashkeel and filter out non-Arabic words:):
        >>> text = u"ِاسمٌ الكلبِ في اللغةِ الإنجليزية Dog واسمُ الحمارِ Donky"
        >>> tokenize(text, conditions=is_arabicrange, morphs=strip_tashkeel)
        ['اسم', 'الكلب', 'في', 'اللغة', 'الإنجليزية', 'واسم', 'الحمار']

    Example 3 (This structure will enable us to create functions on the fly and pass them:):
        >>> text = u"طلع البدر علينا من ثنيات الوداع"
        >>>tokenize(text, conditions=lambda x: x.startswith(u'ال'))
        ['البدر', 'الوداع']

    @param text: the input text.
    @type text: unicode.
    @param conditions: a list of conditions to be applied on tokens, like avoiding non arabic letters.
    @type conditions: one or list of conditions .
    @param morphs: a list of morphological change functions to be applied on tokens, like striping tashkeel or normalizing tokens.
    @type morphs: one or list of morphological functions .
    @return: list of words.
    @rtype: list.
    """
    if text:
        # to be tolerant and allow for a single condition and/or morph to be passed
        # without having to enclose it in a list
        if type(conditions) is not list:
            conditions = [conditions]
        if type(morphs) is not list:
            morphs = [morphs]

        tokens = TOKEN_PATTERN.split(text)
        tokens = [
            TOKEN_REPLACE.sub("", tok) for tok in tokens if TOKEN_REPLACE.sub("", tok)
        ]

        if conditions:
            tokens = [tok for tok in tokens if all([cond(tok) for cond in conditions])]
        if morphs:

            def morph(tok):
                for m in morphs:
                    tok = m(tok)
                return tok

            tokens = [morph(tok) for tok in tokens]
        return tokens
    else:
        return []


def tokenize_with_location(text: str) -> list:
    """
    Tokenize text into words with their positions.

    Example:
        >>> text = "حدثنا ابن أبي عامر، قال: رايت مناما"
        >>> tokens = araby.tokenize_with_location(text)
        >>> print u"\\n".join(tokens)
         [{'token': 'حدثنا', 'start': 0,  'end': 5},
          {'token': 'ابن',   'start': 6,  'end': 9},
          {'token': 'أبي',   'start': 10, 'end': 13},
          {'token': 'عامر',  'start': 14, 'end': 18},
          {'token': 'قال',   'start': 20, 'end': 23},
          {'token': 'رايت',  'start': 25, 'end': 29},
           {'token': 'مناما','start': 30, 'end': 35}
           ]


    @param text: the input text.
    @type text: unicode.
    @return: list of dict of (tokens, starts, ends).
    @rtype: list of dict.
    """
    tokens = []
    for match in TOKEN_PATTERN_SPLIT.finditer(text):
        tokens.append(
            {
                "token": text[match.start() : match.end()],
                "start": match.start(),
                "end": match.end(),
            }
        )

    return tokens


def fix_spaces(text):
    """ """
    text = FIX_SPACES_PAT.sub(lambda x: "{} ".format(x.group(1).replace(" ", "")), text)
    return text.strip()


def autocorrect(text):
    """
    Correct most common errors on word
    like repetetion of harakats,or tanwin befor alef
    @param text: input text
    @type text: unicode
    @return: corrected text
    @rtype: unicode
    """
    ## HARAKAT
    text = re.sub(r"(?<=[\s\d])([%s])+" % (TASHKEEL_STRING), "", text, re.UNICODE)
    text = re.sub("^([%s])+" % (TASHKEEL_STRING), "", text, re.UNICODE)
    # tanwin on alef
    text = re.sub(ALEF + FATHATAN, FATHATAN + ALEF, text, re.UNICODE)

    # SUKUN misplaced on alef /alef maksura and TEH merbuta
    text = re.sub(
        "(?<=[%s%s%s])([%s])+" % (ALEF, ALEF_MAKSURA, TEH_MARBUTA, SUKUN),
        "",
        text,
        re.UNICODE,
    )

    # Hakara before Shadda
    text = re.sub("([%s])+(?=[%s])" % (HARAKAT_STRING, SHADDA), "", text, re.UNICODE)

    # repeated harahat
    text = re.sub(
        "(?<=[%s])([%s])+" % (HARAKAT_STRING, HARAKAT_STRING), "", text, re.UNICODE
    )

    ## Letters
    return text


def spellit(word, lang="ar"):
    """
    write the word in full letter' names
    """

    names = []
    if lang == "unicode":
        for c in word:
            names.append(unicodedata.name(c, c))
    else:
        for c in word:
            names.append(name(c, c))
    return ", ".join(names)


if __name__ == "__main__":
    # ~WORDS = [u'الْدَرَاجَةُ', u'الدّرّاجة',
    # ~u'سّلّامْ', ]
    # ~for wrd in WORDS:
    # ~l, m, s = separate(wrd, True)
    # ~l = joint(l, s)
    # ~print u'\t'.join([wrd, l, m, s]).encode('utf8')
    # ~newword = joint(l, m)
    # ~assert (newword != wrd)
    print("like: ", vocalizedlike("مُتَوَهِّمًا", "متوهمًا"))
    print("sim: ", vocalized_similarity("ثمّ", "ثُمَّ"))
    print("like: ", vocalizedlike("ثمّ", "ثُمَّ"))
    print("sim: ", vocalized_similarity("ثم", "ثُمَّ"))
    print("like: ", vocalizedlike("ثم", "ثُمَّ"))
    print("sim: ", vocalized_similarity("مُتَوَهِّمًا", "متوهمًا"))
    print("sim: ", vocalized_similarity("مُتَوَهِّمًا", "متوهمًا"))
    text1 = "العربية: لغة جميلة."
    wordlist = ["العربية", ":", "لغة", "جميلة", "."]
    wl = tokenize(text1)

    print(" use tokenize")
    print(wl)
    # ~ print((repr(wl)).decode('unicode-escape'))
    # ~ print((repr(wordlist)).decode('unicode-escape'))
    # ~ TOKEN_PATTERN2 = re.compile(u"[^\w\u064b-\u0652']+", re.UNICODE)
    # ~ words = TOKEN_PATTERN2.split(text1)
    # ~ print(" first")
    # ~ print((repr(words)).decode('unicode-escape'))
    # ~ TOKEN_PATTERN3 = re.compile(u"([^\w\u064b-\u0652']+)", re.UNICODE)
    # ~ words = TOKEN_PATTERN3.split(text1)
    # ~ print(" modified")
    # ~ print (repr(words)).decode('unicode-escape')

    # ~ TOKEN_PATTERN4 = re.compile(u"([^\w\u064b-\u0652']+)", re.UNICODE)
    # ~ words = TOKEN_PATTERN4.split(text1)
    # ~ print(" modified without r-prefix")
    # ~ print((repr(words)).decode('unicode-escape'))

    # ~ text = u"ِاسمٌ الكلبِ في اللغةِ الإنجليزية Dog واسمُ الحمارِ Donky"
    # ~ words = tokenize(text, conditions=is_arabicrange, morphs=strip_tashkeel)
    # ~ print((repr(words)).decode('unicode-escape'))

    # ~ #>> ['اسم', 'الكلب', 'في', 'اللغة', 'الإنجليزية', 'واسم', 'الحمار']

    # ~ text = u"طلع البدر علينا من ثنيات الوداع"
    # ~ words = tokenize(text, conditions=lambda x: x.startswith(u'ال'))
    # ~ # >> ['البدر', 'الوداع']
    # ~ print((repr(words)).decode('unicode-escape'))



import os

import requests

from phoonnx.thirdparty.arpa2ipa import arpa_to_ipa_lookup
from phoonnx.phonemizers.base import BasePhonemizer
from phoonnx.config import Alphabet


class DeepPhonemizer(BasePhonemizer):
    """
    https://github.com/spring-media/DeepPhonemizer
    """
    MODELS = {
        "latin_ipa_forward.pt": "https://public-asai-dl-models.s3.eu-central-1.amazonaws.com/DeepPhonemizer/latin_ipa_forward.pt",
        "en_us_cmudict_ipa_forward.pt": "https://public-asai-dl-models.s3.eu-central-1.amazonaws.com/DeepPhonemizer/en_us_cmudict_ipa_forward.pt",
        "en_us_cmudict_forward.pt": "https://public-asai-dl-models.s3.eu-central-1.amazonaws.com/DeepPhonemizer/en_us_cmudict_forward.pt"
    }

    def __init__(self, model="latin_ipa_forward.pt"):
        import dp
        from dp.phonemizer import Phonemizer
        import torch
        # needed for latest torch version
        torch.serialization.add_safe_globals([dp.preprocessing.text.Preprocessor])
        torch.serialization.add_safe_globals([dp.preprocessing.text.LanguageTokenizer])
        torch.serialization.add_safe_globals([dp.preprocessing.text.SequenceTokenizer])

        if "ipa" in model:
            super().__init__(Alphabet.IPA)
        else:
            super().__init__(Alphabet.ARPA)

        if not os.path.isfile(model):
            if model in self.MODELS:
                url = self.MODELS[model]
                cache_dir = os.path.expanduser("~/.local/share/deepphonemizer")
                os.makedirs(cache_dir, exist_ok=True)
                model_path = os.path.join(cache_dir, model)
                if not os.path.isfile(model_path):
                    print(f"Downloading {model} from {url}...")
                    with requests.get(url, stream=True) as r:
                        r.raise_for_status()
                        with open(model_path, 'wb') as f:
                            for chunk in r.iter_content(chunk_size=8192):
                                f.write(chunk)
                    print(f"Saved model to {model_path}")
                model = model_path
            else:
                raise ValueError("invalid model")

        self.phonemizer = Phonemizer.from_checkpoint(model)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ['de', 'en_us'])

    def phonemize_string(self, text: str, lang: str) -> str:
        """
        Normalizes input text by applying a series of transformations
        and returns it as a sequence of graphemes.

        Parameters:
            text (str): Input text to be converted to graphemes.
            lang (str): The language code (ignored for grapheme phonemization,
                        but required by BasePhonemizer).

        Returns:
            str: A normalized string of graphemes.
        """
        lang = self.get_lang(lang)
        return self.phonemizer(text, lang)


class OpenPhonemizer(BasePhonemizer):
    """
    https://github.com/NeuralVox/OpenPhonemizer
    """

    def __init__(self):
        from openphonemizer import OpenPhonemizer
        import torch
        # needed for latest torch version
        import dp
        torch.serialization.add_safe_globals([dp.preprocessing.text.Preprocessor])
        torch.serialization.add_safe_globals([dp.preprocessing.text.LanguageTokenizer])
        torch.serialization.add_safe_globals([dp.preprocessing.text.SequenceTokenizer])

        self.phonemizer = OpenPhonemizer()
        super().__init__(Alphabet.IPA)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["en"])

    def phonemize_string(self, text: str, lang: str) -> str:
        """
        Normalizes input text by applying a series of transformations
        and returns it as a sequence of graphemes.

        Parameters:
            text (str): Input text to be converted to graphemes.
            lang (str): The language code (ignored for grapheme phonemization,
                        but required by BasePhonemizer).

        Returns:
            str: A normalized string of graphemes.
        """
        lang = self.get_lang(lang)
        return self.phonemizer(text)


class G2PEnPhonemizer(BasePhonemizer):
    """
    https://github.com/Kyubyong/g2p
    """

    def __init__(self, alphabet=Alphabet.IPA):
        assert alphabet in [Alphabet.IPA, Alphabet.ARPA]
        import nltk
        nltk.download('averaged_perceptron_tagger_eng')
        nltk.download('cmudict')
        from g2p_en import G2p
        self.g2p = G2p()
        super().__init__(alphabet)

    @classmethod
    def get_lang(cls, target_lang: str) -> str:
        """
        Validates and returns the closest supported language code.

        Args:
            target_lang (str): The language code to validate.

        Returns:
            str: The validated language code.

        Raises:
            ValueError: If the language code is unsupported.
        """
        # this check is here only to throw an exception if invalid language is provided
        return cls.match_lang(target_lang, ["en"])

    def phonemize_string(self, text: str, lang: str) -> str:
        """
        Normalizes input text by applying a series of transformations
        and returns it as a sequence of graphemes.

        Parameters:
            text (str): Input text to be converted to graphemes.
            lang (str): The language code (ignored for grapheme phonemization,
                        but required by BasePhonemizer).

        Returns:
            str: A normalized string of graphemes.
        """
        lang = self.get_lang(lang)
        # NOTE: this model returns ARPA not IPA, may need to map phonemes
        if self.alphabet == Alphabet.ARPA:
            return self.g2p(text)
        return "".join([arpa_to_ipa_lookup.get(pho, pho) for pho in self.g2p(text)])



if __name__ == "__main__":
    # for comparison
    from phoonnx.phonemizers.mul import (ByT5Phonemizer, EspeakPhonemizer, GruutPhonemizer,
                                         EpitranPhonemizer, CharsiuPhonemizer)
    byt5 = ByT5Phonemizer()
    espeak = EspeakPhonemizer()
    gruut = GruutPhonemizer()
    epitr = EpitranPhonemizer()
    charsiu = CharsiuPhonemizer()
    openphon = OpenPhonemizer()
    g2pen = G2PEnPhonemizer()
    dp = DeepPhonemizer()

    lang = "en-gb"

    print("\n--- Getting phonemes for 'Hello, world. How are you?' ---")
    text1 = "Hello, world. How are you?"
    phonemes1 = espeak.phonemize(text1, lang)
    phonemes1b = gruut.phonemize(text1, lang)
    phonemes1c = byt5.phonemize(text1, lang)
    phonemes1d = epitr.phonemize(text1, lang)
    phonemes1e = charsiu.phonemize(text1, lang)
    phonemes1f = openphon.phonemize(text1, lang)
    phonemes1g = g2pen.phonemize(text1, lang)
    phonemes1h = dp.phonemize(text1, lang)
    print(f" Espeak         Phonemes: {phonemes1}")
    print(f" Gruut          Phonemes: {phonemes1b}")
    print(f" byt5           Phonemes: {phonemes1c}")
    print(f" Epitran        Phonemes: {phonemes1d}")
    print(f" Charsiu        Phonemes: {phonemes1e}")
    print(f" OpenPhonemizer Phonemes: {phonemes1f}")
    print(f" DeepPhonemizer Phonemes: {phonemes1h}")
    print(f" G2P_en         Phonemes: {phonemes1g}")

    print("\n--- Getting phonemes for 'This is a test: a quick one; and done!' ---")
    text2 = "This is a test: a quick one; and done!"
    phonemes2 = espeak.phonemize(text2, lang)
    phonemes2b = gruut.phonemize(text2, lang)
    phonemes2c = byt5.phonemize(text2, lang)
    phonemes2d = epitr.phonemize(text2, lang)
    phonemes2e = charsiu.phonemize(text2, lang)
    print(f"  Espeak Phonemes: {phonemes2}")
    print(f"  Gruut Phonemes: {phonemes2b}")
    print(f"   byt5  Phonemes: {phonemes2c}")
    print(f" Epitran Phonemes: {phonemes2d}")
    print(f" Charsiu Phonemes: {phonemes2e}")

    print("\n--- Getting phonemes for 'Just a phrase without punctuation' ---")
    text3 = "Just a phrase without punctuation"
    phonemes3 = espeak.phonemize(text3, lang)
    phonemes3b = gruut.phonemize(text3, lang)
    phonemes3c = byt5.phonemize(text3, lang)
    phonemes3d = epitr.phonemize(text3, lang)
    phonemes3e = charsiu.phonemize(text3, lang)
    print(f"  Espeak Phonemes: {phonemes3}")
    print(f"  Gruut Phonemes: {phonemes3b}")
    print(f"   byt5  Phonemes: {phonemes3c}")
    print(f" Epitran Phonemes: {phonemes3d}")
    print(f" Charsiu Phonemes: {phonemes3e}")




PADDING_TOKEN = "_pad_"
EOS_TOKEN = "_eos_"
DOUBLING_TOKEN = "_dbl_"
SEPARATOR_TOKEN = "_+_"

EOS_TOKENS = [SEPARATOR_TOKEN, EOS_TOKEN]

symbols = [
    # special tokens
    PADDING_TOKEN,  # padding
    EOS_TOKEN,  # eos-token
    "_sil_",  # silence
    DOUBLING_TOKEN,  # doubling
    SEPARATOR_TOKEN,  # word separator
    # punctuation
    ".",
    ",",
    "?",
    "!",
    ":",
    ";",
    "-",
    ")",
    "(",
    '"',
    # consonants
    "<",  # hamza
    "b",  # baa'
    "t",  # taa'
    "^",  # thaa'
    "j",  # jiim
    "H",  # Haa'
    "x",  # xaa'
    "d",  # daal
    "*",  # dhaal
    "r",  # raa'
    "z",  # zaay
    "s",  # siin
    "$",  # shiin
    "S",  # Saad
    "D",  # Daad
    "T",  # Taa'
    "Z",  # Zhaa'
    "E",  # 3ayn
    "g",  # ghain
    "f",  # faa'
    "q",  # qaaf
    "k",  # kaaf
    "l",  # laam
    "m",  # miim
    "n",  # nuun
    "h",  # haa'
    "w",  # waaw
    "y",  # yaa'
    "v",  # /v/ for loanwords e.g. in u'fydyw': u'v i0 d y uu1',
    # vowels
    "a",  # short
    "u",
    "i",
    "aa",  # long
    "uu",
    "ii",
    "aaaa",
]



#!/usr/bin/python
# -*- coding: UTF8 -*-

# adapted from: https://github.com/nawarhalabi/Arabic-Phonetiser/blob/master/phonetise-Buckwalter.py
# license: Creative Commons Attribution-NonCommercial 4.0 International License.
# https://creativecommons.org/licenses/by-nc/4.0/

import re

arabic_to_buckw_dict = {  # mapping from Arabic script to Buckwalter
    "\u0628": "b",
    "\u0630": "*",
    "\u0637": "T",
    "\u0645": "m",
    "\u062a": "t",
    "\u0631": "r",
    "\u0638": "Z",
    "\u0646": "n",
    "\u062b": "^",
    "\u0632": "z",
    "\u0639": "E",
    "\u0647": "h",
    "\u062c": "j",
    "\u0633": "s",
    "\u063a": "g",
    "\u062d": "H",
    "\u0642": "q",
    "\u0641": "f",
    "\u062e": "x",
    "\u0635": "S",
    "\u0634": "$",
    "\u062f": "d",
    "\u0636": "D",
    "\u0643": "k",
    "\u0623": ">",
    "\u0621": "'",
    "\u0626": "}",
    "\u0624": "&",
    "\u0625": "<",
    "\u0622": "|",
    "\u0627": "A",
    "\u0649": "Y",
    "\u0629": "p",
    "\u064a": "y",
    "\u0644": "l",
    "\u0648": "w",
    "\u064b": "F",
    "\u064c": "N",
    "\u064d": "K",
    "\u064e": "a",
    "\u064f": "u",
    "\u0650": "i",
    "\u0651": "~",
    "\u0652": "o",
}

buckw_to_arabic_dict = {  # mapping from Buckwalter to Arabic script
    "b": "\u0628",
    "*": "\u0630",
    "T": "\u0637",
    "m": "\u0645",
    "t": "\u062a",
    "r": "\u0631",
    "Z": "\u0638",
    "n": "\u0646",
    "^": "\u062b",
    "z": "\u0632",
    "E": "\u0639",
    "h": "\u0647",
    "j": "\u062c",
    "s": "\u0633",
    "g": "\u063a",
    "H": "\u062d",
    "q": "\u0642",
    "f": "\u0641",
    "x": "\u062e",
    "S": "\u0635",
    "$": "\u0634",
    "d": "\u062f",
    "D": "\u0636",
    "k": "\u0643",
    ">": "\u0623",
    "'": "\u0621",
    "}": "\u0626",
    "&": "\u0624",
    "<": "\u0625",
    "|": "\u0622",
    "A": "\u0627",
    "Y": "\u0649",
    "p": "\u0629",
    "y": "\u064a",
    "l": "\u0644",
    "w": "\u0648",
    "F": "\u064b",
    "N": "\u064c",
    "K": "\u064d",
    "a": "\u064e",
    "u": "\u064f",
    "i": "\u0650",
    "~": "\u0651",
    "o": "\u0652",
}


def arabic_to_buckwalter(word):  # Convert input string to Buckwalter
    res = ""
    for letter in word:
        if letter in arabic_to_buckw_dict:
            res += arabic_to_buckw_dict[letter]
        else:
            res += letter
    return res


def buckwalter_to_arabic(word):  # Convert input string to Arabic
    res = ""
    for letter in word:
        if letter in buckw_to_arabic_dict:
            res += buckw_to_arabic_dict[letter]
        else:
            res += letter
    return res


# ----------------------------------------------------------------------------
# Grapheme to Phoneme mappings------------------------------------------------
# ----------------------------------------------------------------------------
unambiguousConsonantMap = {
    "b": "b",
    "*": "*",
    "T": "T",
    "m": "m",
    "t": "t",
    "r": "r",
    "Z": "Z",
    "n": "n",
    "^": "^",
    "z": "z",
    "E": "E",
    "h": "h",
    "j": "j",
    "s": "s",
    "g": "g",
    "H": "H",
    "q": "q",
    "f": "f",
    "x": "x",
    "S": "S",
    "$": "$",
    "d": "d",
    "D": "D",
    "k": "k",
    ">": "<",
    "'": "<",
    "}": "<",
    "&": "<",
    "<": "<",
}

ambiguousConsonantMap = {
    # These consonants are only unambiguous in certain contexts
    "l": ["l", ""],
    "w": "w",
    "y": "y",
    "p": ["t", ""],
}

maddaMap = {"|": [["<", "aa"], ["<", "AA"]]}

vowelMap = {
    "A": [["aa", ""], ["AA", ""]],
    "Y": [["aa", ""], ["AA", ""]],
    "w": [["uu0", "uu1"], ["UU0", "UU1"]],
    "y": [["ii0", "ii1"], ["II0", "II1"]],
    "a": ["a", "A"],
    "u": [["u0", "u1"], ["U0", "U1"]],
    "i": [["i0", "i1"], ["I0", "I1"]],
}

nunationMap = {
    "F": [["a", "n"], ["A", "n"]],
    "N": [["u1", "n"], ["U1", "n"]],
    "K": [["i1", "n"], ["I1", "n"]],
}

diacritics = ["o", "a", "u", "i", "F", "N", "K", "~"]
diacriticsWithoutShadda = ["o", "a", "u", "i", "F", "N", "K"]
emphatics = ["D", "S", "T", "Z", "g", "x", "q"]
forwardEmphatics = ["g", "x"]
consonants = [
    ">",
    "<",
    "}",
    "&",
    "'",
    "b",
    "t",
    "^",
    "j",
    "H",
    "x",
    "d",
    "*",
    "r",
    "z",
    "s",
    "$",
    "S",
    "D",
    "T",
    "Z",
    "E",
    "g",
    "f",
    "q",
    "k",
    "l",
    "m",
    "n",
    "h",
    "|",
]

punctuation = [".", ",", "?", "!"]

# ------------------------------------------------------------------------------------
# Words with fixed irregular pronunciations-------------------------------------------
# ------------------------------------------------------------------------------------
fixedWords = {
    "h*A": [
        "h aa * aa",
        "h aa * a",
    ],
    "h*h": ["h aa * i0 h i0", "h aa * i1 h"],
    "h*An": ["h aa * aa n i0", "h aa * aa n"],
    "h&lA'": ["h aa < u0 l aa < i0", "h aa < u0 l aa <"],
    "*lk": ["* aa l i0 k a", "* aa l i0 k"],
    "k*lk": ["k a * aa l i0 k a", "k a * aa l i1 k"],
    "*lkm": "* aa l i0 k u1 m",
    ">wl}k": ["< u0 l aa < i0 k a", "< u0 l aa < i1 k"],
    "Th": "T aa h a",
    "lkn": ["l aa k i0 nn a", "l aa k i1 n"],
    "lknh": "l aa k i0 nn a h u0",
    "lknhm": "l aa k i0 nn a h u1 m",
    "lknk": ["l aa k i0 nn a k a", "l aa k i0 nn a k i0"],
    "lknkm": "l aa k i0 nn a k u1 m",
    "lknkmA": "l aa k i0 nn a k u0 m aa",
    "lknnA": "l aa k i0 nn a n aa",
    "AlrHmn": ["rr a H m aa n i0", "rr a H m aa n"],
    "Allh": ["ll aa h i0", "ll aa h", "ll AA h u0", "ll AA h a", "ll AA h", "ll A"],
    "h*yn": ["h aa * a y n i0", "h aa * a y n"],
    "nt": "n i1 t",
    "fydyw": "v i0 d y uu1",
    "lndn": "l A n d u1 n",
}


def isFixedWord(word, results, orthography, pronunciations):
    lastLetter = ""
    if len(word) > 0:
        lastLetter = word[-1]
    if lastLetter == "a":
        lastLetter = ["a", "A"]
    elif lastLetter == "A":
        lastLetter = ["aa"]
    elif lastLetter == "u":
        lastLetter = ["u0"]
    elif lastLetter == "i":
        lastLetter = ["i0"]
    elif lastLetter in unambiguousConsonantMap:
        lastLetter = [unambiguousConsonantMap[lastLetter]]
    # Remove all dacritics from word
    wordConsonants = re.sub(r"[^h*Ahn\'>wl}kmyTtfd]", "", word)
    if wordConsonants in fixedWords:  # check if word is in the fixed word lookup table
        if isinstance(fixedWords[wordConsonants], list):
            for pronunciation in fixedWords[wordConsonants]:
                if pronunciation.split(" ")[-1] in lastLetter:
                    # add each pronunciation to the pronunciation dictionary
                    results += word + " " + pronunciation + "\n"
                    pronunciations.append(pronunciation.split(" "))
        else:
            # add pronunciation to the pronunciation dictionary
            results += word + " " + fixedWords[wordConsonants] + "\n"
            pronunciations.append(fixedWords[wordConsonants].split(" "))
    return results


def preprocess_utterance(utterance):
    # Do some normalisation work and split utterance to words
    utterance = utterance.replace("AF", "F")
    utterance = utterance.replace("\u0640", "")
    utterance = utterance.replace("o", "")
    utterance = utterance.replace("aA", "A")
    utterance = utterance.replace("aY", "Y")
    utterance = utterance.replace(" A", " ")
    utterance = utterance.replace("F", "an")
    utterance = utterance.replace("N", "un")
    utterance = utterance.replace("K", "in")
    utterance = utterance.replace("|", ">A")

    utterance = utterance.replace("i~", "~i")
    utterance = utterance.replace("a~", "~a")
    utterance = utterance.replace("u~", "~u")

    # Deal with Hamza types that when not followed by a short vowel letter,
    # this short vowel is added automatically
    utterance = re.sub(r"Ai", "<i", utterance)
    utterance = re.sub(r"Aa", ">a", utterance)
    utterance = re.sub(r"Au", ">u", utterance)
    utterance = re.sub(r"^>([^auAw])", ">a\\1", utterance)
    utterance = re.sub(r" >([^auAw ])", " >a\\1", utterance)
    utterance = re.sub(r"<([^i])", "<i\\1", utterance)

    utterance = re.sub(r"(\S)(\.|\?|,|!)", "\\1 \\2", utterance)

    utterance = utterance.split(" ")

    return utterance


def process_word(word):

    if word in punctuation:
        return word

    pronunciations = (
        []
    )  # Start with empty set of possible pronunciations of current word
    # Add fixed irregular pronuncations if possible
    isFixedWord(word, "", word, pronunciations)

    # Indicates whether current character is in an emphatic context or not. Starts with False
    emphaticContext = False
    # This is the end/beginning of word symbol. just for convenience
    word = "bb" + word + "ee"

    phones = []  # Empty list which will hold individual possible word's pronunciation

    # -----------------------------------------------------------------------------------
    # MAIN LOOP: here is where the Modern Standard Arabic phonetisation rule-set starts--
    # -----------------------------------------------------------------------------------
    for index in range(2, len(word) - 2):
        letter = word[index]  # Current Character
        letter1 = word[index + 1]  # Next Character
        letter2 = word[index + 2]  # Next-Next Character
        letter_1 = word[index - 1]  # Previous Character
        letter_2 = word[index - 2]  # Before Previous Character
        # ----------------------------------------------------------------------------------------------------------------
        if letter in consonants + ["w", "y"] and not letter in emphatics + [
            "r" """, u'l'"""
        ]:  # non-emphatic consonants (except for Lam and Ra) change emphasis back to False
            emphaticContext = False
        if letter in emphatics:  # Emphatic consonants change emphasis context to True
            emphaticContext = True
        # If following letter is backward emphatic, emphasis state is set to True
        if letter1 in emphatics and not letter1 in forwardEmphatics:
            emphaticContext = True
        # ----------------------------------------------------------------------------------------------------------------
        # ----------------------------------------------------------------------------------------------------------------
        # Unambiguous consonant phones. These map to a predetermined phoneme
        if letter in unambiguousConsonantMap:
            phones += [unambiguousConsonantMap[letter]]
        # ----------------------------------------------------------------------------------------------------------------
        if letter == "l":  # Lam is a consonant which requires special treatment
            # Lam could be omitted in definite article (sun letters)
            if (not letter1 in diacritics and not letter1 in vowelMap) and letter2 in [
                "~"
            ]:
                phones += [ambiguousConsonantMap["l"][1]]  # omit
            else:
                # do not omit
                phones += [ambiguousConsonantMap["l"][0]]
        # ----------------------------------------------------------------------------------------------------------------
        # shadda just doubles the letter before it
        if letter == "~" and not letter_1 in ["w", "y"] and len(phones) > 0:
            phones[-1] += phones[-1]
        # ----------------------------------------------------------------------------------------------------------------
        if letter == "|":  # Madda only changes based in emphaticness
            if emphaticContext:
                phones += [maddaMap["|"][1]]
            else:
                phones += [maddaMap["|"][0]]
        # ----------------------------------------------------------------------------------------------------------------
        if (
            letter == "p"
        ):  # Ta' marboota is determined by the following if it is a diacritic or not
            if letter1 in diacritics:
                phones += [ambiguousConsonantMap["p"][0]]
            else:
                phones += [ambiguousConsonantMap["p"][1]]
        # ----------------------------------------------------------------------------------------------------------------
        if letter in vowelMap:
            # Waw and Ya are complex they could be consonants or vowels and their gemination is complex as it could be a combination of a vowel and consonants
            if letter in ["w", "y"]:
                if (
                    letter1 in diacriticsWithoutShadda + ["A", "Y"]
                    or (
                        letter1 in ["w", "y"]
                        and not letter2 in diacritics + ["A", "w", "y"]
                    )
                    or (
                        letter_1 in diacriticsWithoutShadda
                        and letter1 in consonants + ["e"]
                    )
                ):
                    if (
                        letter in ["w"]
                        and letter_1 in ["u"]
                        and not letter1 in ["a", "i", "A", "Y"]
                    ) or (
                        letter in ["y"]
                        and letter_1 in ["i"]
                        and not letter1 in ["a", "u", "A", "Y"]
                    ):
                        if emphaticContext:
                            phones += [vowelMap[letter][1][0]]
                        else:
                            phones += [vowelMap[letter][0][0]]
                    else:
                        if letter1 in ["A"] and letter in ["w"] and letter2 in ["e"]:
                            phones += [
                                [ambiguousConsonantMap[letter], vowelMap[letter][0][0]]
                            ]
                        else:
                            phones += [ambiguousConsonantMap[letter]]
                elif letter1 in ["~"]:
                    if (
                        letter_1 in ["a"]
                        or (letter in ["w"] and letter_1 in ["i", "y"])
                        or (letter in ["y"] and letter_1 in ["w", "u"])
                    ):
                        phones += [
                            ambiguousConsonantMap[letter],
                            ambiguousConsonantMap[letter],
                        ]
                    else:
                        phones += [
                            vowelMap[letter][0][0],
                            ambiguousConsonantMap[letter],
                        ]
                else:  # Waws and Ya's at the end of the word could be shortened
                    if emphaticContext:
                        if letter_1 in consonants + ["u", "i"] and letter1 in ["e"]:
                            phones += [
                                [vowelMap[letter][1][0], vowelMap[letter][1][0][1:]]
                            ]
                        else:
                            phones += [vowelMap[letter][1][0]]
                    else:
                        if letter_1 in consonants + ["u", "i"] and letter1 in ["e"]:
                            phones += [
                                [vowelMap[letter][0][0], vowelMap[letter][0][0][1:]]
                            ]
                        else:
                            phones += [vowelMap[letter][0][0]]
            # Kasra and Damma could be mildened if before a final silent consonant
            if letter in ["u", "i"]:
                if emphaticContext:
                    if (
                        (letter1 in unambiguousConsonantMap or letter1 == "l")
                        and letter2 == "e"
                        and len(word) > 7
                    ):
                        phones += [vowelMap[letter][1][1]]
                    else:
                        phones += [vowelMap[letter][1][0]]
                else:
                    if (
                        (letter1 in unambiguousConsonantMap or letter1 == "l")
                        and letter2 == "e"
                        and len(word) > 7
                    ):
                        phones += [vowelMap[letter][0][1]]
                    else:
                        phones += [vowelMap[letter][0][0]]
            # Alif could be ommited in definite article and beginning of some words
            if letter in ["a", "A", "Y"]:
                if letter in ["A"] and letter_1 in ["w", "k"] and letter_2 == "b":
                    phones += [["a", vowelMap[letter][0][0]]]
                elif letter in ["A"] and letter_1 in ["u", "i"]:
                    temp = True  # do nothing
                # Waw al jama3a: The Alif after is optional
                elif letter in ["A"] and letter_1 in ["w"] and letter1 in ["e"]:
                    phones += [[vowelMap[letter][0][0], vowelMap[letter][0][1]]]
                elif letter in ["A", "Y"] and letter1 in ["e"]:
                    if emphaticContext:
                        phones += [[vowelMap[letter][1][0], vowelMap["a"][1]]]
                    else:
                        phones += [[vowelMap[letter][0][0], vowelMap["a"][0]]]
                else:
                    if emphaticContext:
                        phones += [vowelMap[letter][1][0]]
                    else:
                        phones += [vowelMap[letter][0][0]]
    # -------------------------------------------------------------------------------------------------------------------------
    # End of main loop---------------------------------------------------------------------------------------------------------
    # -------------------------------------------------------------------------------------------------------------------------
    possibilities = 1  # Holds the number of possible pronunciations of a word

    # count the number of possible pronunciations
    for letter in phones:
        if isinstance(letter, list):
            possibilities = possibilities * len(letter)

    # Generate all possible pronunciations
    for i in range(0, possibilities):
        pronunciations.append([])
        iterations = 1
        for index, letter in enumerate(phones):
            if isinstance(letter, list):
                curIndex = int(i / iterations) % len(letter)
                if letter[curIndex] != "":
                    pronunciations[-1].append(letter[curIndex])
                iterations = iterations * len(letter)
            else:
                if letter != "":
                    pronunciations[-1].append(letter)

    # Iterate through each pronunciation to perform some house keeping. And append pronunciation to dictionary
    # 1- Remove duplicate vowels
    # 2- Remove duplicate y and w
    for pronunciation in pronunciations:
        prevLetter = ""
        toDelete = []
        for i in range(0, len(pronunciation)):
            letter = pronunciation[i]
            # Delete duplicate consecutive vowels
            if (
                letter in ["aa", "uu0", "ii0", "AA", "UU0", "II0"]
                and prevLetter.lower() == letter[1:].lower()
            ):
                toDelete.append(i - 1)
                pronunciation[i] = pronunciation[i - 1][0] + pronunciation[i - 1]
            # Delete duplicates
            if letter in ["u0", "i0"] and prevLetter.lower() == letter.lower():
                toDelete.append(i - 1)
                pronunciation[i] = pronunciation[i - 1]
            if letter in ["y", "w"] and prevLetter == letter:  # delete duplicate
                pronunciation[i - 1] += pronunciation[i - 1]
                toDelete.append(i)

            prevLetter = letter
        for i in reversed(range(0, len(toDelete))):
            del pronunciation[toDelete[i]]

    return pronunciations[0]


def process_utterance(utterance):

    utterance = preprocess_utterance(utterance)
    phonemes = []

    for word in utterance:
        if word in ["-", "sil"]:
            phonemes.append(["sil"])
            continue

        phonemes_word = process_word(word)
        if phonemes_word in punctuation and phonemes:
            phonemes[-1] += phonemes_word
        else:
            phonemes.append(phonemes_word)

    final_sequence = " + ".join(
        " ".join(phon for phon in phones) for phones in phonemes
    )

    return final_sequence



#!/usr/bin/python
# -*- coding: UTF-8 -*-
"""
Improve repr predifined function to best display of objects containing unicode
Unicode represention texts
@author: Taha Zerrouki
@contact: taha dot zerrouki at gmail dot com
@copyright: Taha Zerrouki
@license: GPL
@date:2014/03/01
@version: 0.1
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sys

if sys.version_info < (3, 0):
    import repr as reprlib
else:
    import reprlib


class ArabicRepr(reprlib.Repr):
    """A redifinition of repr fucntion,
    you can use it like this

    Example:
        >>> import pyarabic.arabrepr as arabrepr
        >>> arepr = arabrepr.ArabicRepr()
        >>> repr = arepr.repr
        >>> word = u"السلام عليكم ورحمة الله"
        >>> wordlist = word.split(" ")
        >>> print wordlist
        [u'\u0627\u0644\u0633\u0644\u0627\u0645',
        u'\u0639\u0644\u064a\u0643\u0645',
        u'\u0648\u0631\u062d\u0645\u0629',
        u'\u0627\u0644\u0644\u0647']
        >>> print repr(wordlist)
        [u'السلام', u'عليكم', u'ورحمة', u'الله']
    """

    def repr_unicode(self, obj, level):
        "Modify unicode display"
        return "u'%s'" % obj

    def arepr_unicode(self, obj, level):
        "Modify unicode display"
        return "u'%s'" % obj


if sys.version_info < (3, 0):

    def arepr(data):
        """display a dict with arabic text properly"""
        return repr(data).replace("},", "},\n").decode("unicode-escape").encode("utf8")

else:

    def arepr(data):
        """display a dict with arabic text properly"""
        return repr(data).replace(
            "},", "},\n"
        )  # .decode('unicode-escape').encode('utf8')



from .phonetise_buckwalter import (arabic_to_buckwalter, buckwalter_to_arabic,
                                   process_utterance)
from .symbols import DOUBLING_TOKEN, EOS_TOKEN, SEPARATOR_TOKEN, symbols

vowels = [
    "aa",
    "AA",
    "uu0",
    "uu1",
    "UU0",
    "UU1",
    "ii0",
    "ii1",
    "II0",
    "II1",
    "a",
    "A",
    "u0",
    "u1",
    "U0",
    "U1",
    "i0",
    "i1",
    "I0",
    "I1",
]

vowel_map = {
    "aa": "aa",
    "AA": "aa",
    "uu0": "uu",
    "uu1": "uu",
    "UU0": "uu",
    "UU1": "uu",
    "ii0": "ii",
    "ii1": "ii",
    "II0": "ii",
    "II1": "ii",
    "a": "a",
    "A": "a",
    "u0": "u",
    "u1": "u",
    "U0": "u",
    "U1": "u",
    "i0": "i",
    "i1": "i",
    "I0": "i",
    "I1": "i",
}

phon_to_id_ = {phon: i for i, phon in enumerate(symbols)}


def tokens_to_ids(phonemes, phon_to_id=None):
    if phon_to_id is None:
        return [phon_to_id_[phon] for phon in phonemes]
    return [phon_to_id[phon] for phon in phonemes]


def ids_to_tokens(ids):
    return [symbols[id] for id in ids]


def arabic_to_phonemes(arabic):
    buckw = arabic_to_buckwalter(arabic)
    return process_utterance(buckw)


def buckwalter_to_phonemes(buckw):
    return process_utterance(buckw)


def phonemes_to_tokens(phonemes: str, append_space=False):
    phonemes = phonemes.replace("sil", "").replace("+", "_+_").split()
    for i, phon in enumerate(phonemes):
        if len(phon) == 2 and phon not in vowels and phon[0] == phon[1]:
            phonemes[i] = phon[0]
            phonemes.insert(i + 1, DOUBLING_TOKEN)
        if phonemes[i] in vowels:
            phonemes[i] = vowel_map[phonemes[i]]

    if append_space:
        phonemes.append(SEPARATOR_TOKEN)

    phonemes.append(EOS_TOKEN)

    return phonemes


def buckwalter_to_tokens(buckw, append_space=False):
    phonemes = buckwalter_to_phonemes(buckw)
    tokens = phonemes_to_tokens(phonemes, append_space=append_space)
    return tokens


def arabic_to_tokens(arabic, append_space=False):
    buckw = arabic_to_buckwalter(arabic)
    tokens = buckwalter_to_tokens(buckw, append_space=append_space)
    return tokens


def simplify_phonemes(phonemes):
    for k, v in vowel_map.items():
        phonemes = phonemes.replace(k, v)
    return phonemes
